/*************************************************************************************************
 *                         PVCS Version Control Information                                      *
 *                                                                                               *
 * $Header:   P:\d878\HomeLAN\Archives\NdisDriver\Mii.c_v   1.2   08 Sep 1999 13:22:38   wangmx  $                            
 *                                                                                               *
 ************************************************************************************************/

/*
 *  Module Name: mii.c
 *  Author: Joe Bonaker
 *
 *  Copyright  1997-1998 Conexant Systems, Inc.
 *  All rights reserved.
 *
 *  This software may be used and distributed according to the terms
 *  of the GNU Public License, incorporated herein
 */

/*
 *  Module Description: 100Base-T Media Independent Interface (MII)
 *  module, which contains adapter and OS-specific support routines for
 *  the "phy.c" PHY driver module.  This file must be tailored for each
 *  unique adapter/operating system environment pair.
 *
 *  Conditional Compilation Switches:
 *
 *     The following conditional compilation switches apply, and may
 *     be overridden by the user at compile-time.
 *
 *     DRV_OS - default DOS.  Establishes the OS-specific controller
 *     (MAC or repeater) device I/O (or memory-mapped) access to the
 *     PHY's serial MII plus microsecond and/or millisecond timer
 *     support as detailed in [Step 1] of the Porting Guide below and
 *     within this module.  DRV_OS types are defined in "phytypes.h."
 *
 *     PHY_HWRESET() - default #defined as the MiiPhyHwReset() function
 *     contained within this module.  This can be overridden by
 *     #defining a null or alternate, user-supplied PHY hardware reset
 *     routine if necessary.
 *
 *     The GPIO_* macros control the behavior of a GPIO-based hardware
 *     reset optionally issued when the "phy.c" PhyInit() routine
 *     optionally calls PHY_HWRESET() based on selection of one of the
 *     PHY_HWOPTS_RESET_* hardware option masks.
 *
 *     GPIO_DEFAULT_CTRL - default #undefined (none).  Defines
 *     additional GPIO lines (besides the PHY HW reset GPIO line) to be
 *     configured as output signals.  In the default case, only the
 *     GPIO line used for PHY HW reset is configured as an output GPIO
 *     signal; the other three GPIO signals are configured as inputs.
 *     The GPIO_DEFAULT_CTRL format is per MAC register 15 (R15), where
 *     GP[3:0] are respectively mapped to R15 bits [19:16].
 *
 *     GPIO_DEFAULT_DATA - default #undefined.  Has no effect unless
 *     GPIO_DEFAULT_CTRL is also defined.  When #undefined, the GPIO
 *     register (R15 on RSS 11617 and newer MAC devices, R12 on
 *     pre-production RSS 11611, 11610 and 11606 devices) is read
 *     prior to setting PHY hardware GPIO line.  In this manner, the
 *     state of a previously issued GPIO output data command is
 *     retained subsequent to PHY hardware reset.  When #defined, the
 *     value of GPIO_DEFAULT_DATA is forced out the non-PHY HW reset
 *     GPIO output signals designated by GPIO_DEFAULT_CTRL.  The bit
 *     format is per MAC R15 convention, where GP[3:0] are respectively
 *     mapped to R15 bits [19:16].
 *
 *     GPIO_INTERWORD_GAP - default 1 (microsecond).  Effectively
 *     defines the width of the PHY hardware reset pulse generated
 *     on one of the GP[3:0] lines.  Some PHY devices may require
 *     a wider reset pulse, thus this value can be increased
 *     with microsecond granularity if necessary.
 *
 *     GPIO_SETTLE_TIME - default 1 (millescond).  Provides a delay
 *     between the hardware reset pulse and serial MII interrogation
 *     and control of the PHY device.  Some PHY devices may require a
 *     longer delay, therefore this value can be increased with
 *     millisecond granularity if necessary.
 *
 *  Porting Guide:
 *
 *  There are two steps which must be performed to tailor this module
 *  for a specific adapter (NIC) or repeater configuration.
 *
 *  [Step 1] entails mapping the OS_MICROSECOND_WAIT() and/or
 *  OS_MILLISECOND_WAIT(), IO_WRITE(), IO_READ(), plus optional
 *  OS_NANOSECOND_WAIT() macros into the PHY driver's operating system,
 *  as defined by DRV_OS.
 *
 *  [Step 2] alters the adapter-specific register configuration.  At
 *  present, only a MAC (NIC) controller interface, the RSS1161x 10/100
 *  Mbit Fast Ethernet series, is supported.  [Step 2a] contains
 *  MAC general purpose I/O (GPIO) settings which may vary among

 *  different adapter configurations.  [Step 2b] contains an optional
 *  PHY hardware reset sequence, which is dependent upon the [Step 2a]
 *  interconnection between the MAC and PHY devices/modules.
 *
 *  Search this module for occurances of [Step 1] and [Step 2], editing
 *  accordingly.
 *
 * $Name: v2.07 $
 * $Log:   P:\d878\HomeLAN\Archives\NdisDriver\Mii.c_v  $
 * 
 *    Rev 1.2   08 Sep 1999 13:22:38   wangmx
 * JC:
 * Fix problem which prevent HCF from reading
 * EEPROM when HLAN is loaded.
 * 
 *    Rev 1.1   May 12 1999 12:05:56   wangmx
 *  
 * Revision 1.17  1998/10/07 22:09:23  bonakeaj
 * - Integrated 1M8 HomePNA with multi-PHY support
 * Revision 1.16  1998/02/12 22:54:06  bonakeaj
 * - RSS 11617 R15 GPIO PHY_HWRESET(), WIN32_DIAG NT 4.0 merge
 * Revision 1.15  1998/02/05 01:06:30  bonakeaj
 * - Preamble supression,  QSI, LXT, GEC, TDK and ICS reg definition
 * Revision 1.14  1998/02/02 20:10:38  bonakeaj
 * - Borland C/C++ NDIS 2.01 (DRV_OS == DOS) support via tools.ide
 * Revision 1.13  1998/01/21 23:22:00  bonakeaj
 * - Davicom PHY and DRV_OS=SCO_MDI support
 * Revision 1.12  1998/01/08 14:58:26  bonakeaj
 * - Merged WIN32_DIAG support for NT 4.0
 * Revision 1.11  1997/11/03 21:05:56  bonakeaj
 * - WIN32_DIAG GetCSR/PutCSR now bracketed by open/close driver
 * Revision 1.10  1997/09/16 19:08:40  bonakeaj
 * - SAR-55, PHY HW reset was-500 is-1 mSec, exclusive HW/SW reset only
 * Revision 1.9  1997/09/12 00:09:52  bonakeaj
 * - DRV_OS=LINUX port
 * Revision 1.8  1997/08/21 21:54:42  kamstrd
 * - Updated copyright headers
 * - Replaced RW116?? with RSS116??
 * - Replaced RW11606 with RSS1161x in some comments
 * Revision 1.7  1997/08/21 21:16:38  bonakeaj
 * - DRV_OS=NOVELL_CHSM changes (minor) to support Novell
 * - CHSM-32 driver integration
 * Revision 1.6  1997/07/24 16:32:36  bonakeaj
 * - Timing and NSC DP83840 changes for WIN32_DIAG integration
 * Revision 1.5  1997/07/21 22:42:48  bonakeaj
 * - Fixed OS_NANSECOND_WAIT() code bug -- was uSec, not nSec
 * Revision 1.4  1997/07/17 22:27:54  bonakeaj
 * - Added DRV_OS=WIN32_DIAG, and PHY_CONN_ANY_* support
 * Revision 1.3  1997/07/12 02:23:54  bonakeaj
 * - Added PHY_LINK_NEW_MEDIA support, minor type changes for
 * - NDIS 3.1 macApi/phyDrv integration, and fixed PhyInit() bug where
 * - PhyDisable()'s isolate action prevented Auto-Negotiation
 * - completion with the NSC DP83840 PHY device.
 **********************************************************************/

#include "phy.h"
#include "mii.h"

 /*--------------------------------------------------------------------*
 * [Step 1] Enter DRV_OS-dependent support for OS_MICROSECOND_WAIT(),
 *          and/or OS_MILLISECOND_WAIT(), optional OS_NANOSECOND_WAIT(),
 *          plus 32-bit IO_WRITE()/IO_READ() services.  Edit one or
 *          more of [Step 1a] through [Step 1h] as necessary.
 *
 *          Either one or both of OS_MICROSECOND_WAIT() and
 *          OS_MILLISECOND_WAIT() must be #defined, depending on DRV_OS
 *          environment support.  If the DRV_OS supports microsecond
 *          granularity delays (preferred), then #define the
 *          OS_MICROSECOND_WAIT() macro accordingly.  Otherwise,
 *          #define OS_MILLISECOND_WAIT().  If the DRV_OS supports both
 *          microsecond and millisecond delays, and advocates a scheme
 *          for relinquishing the CPU during longer delays, then
 *          OS_MILLISECOND_WAIT() may also be #defined using the more
 *          system-friendly long delay service.  As long as at least
 *          one of these two macros is #defined, code outside this
 *          [Step 1] DRV_OS-specific block will ensure that the other
 *          macro is properly derived.
 *
 *          Unless the DRV_OS supports nanosecond granularity (highly
 *          unlikely), OS_NANOSECOND_WAIT() should NOT be #defined,
 *          and will be derived from OS_MICROSECOND_WAIT() outside this
 *          code block.  There is one exception.  Since the nanosecond
 *          delay is only used to ensure that MiiRead() and MiiWrite()
 *          clock timing doesn't exceed the 802.3u minimum hold time
 *          of 200-300 nanoseconds, OS_NANOSECOND_WAIT() can be
 *          disabled in a DRV_OS environment where IO_READ() and
 *          IO_WRITE() take a considerable length of time to complete.
 *          This is accomplished as follows:
 *
 *              #define OS_NANOSECOND_WAIT(n)           // No delay
 *
 *          For example, the DRV_OS=WIN32_DIAG relies on driver ioctls
 *          to perform IO_READ() and IO_WRITE().  This, in conjunction
 *          with a 10-millisecond delay granularity, results in
 *          MiiRead() and MiiWrite() times approaching one second per
 *          register!  Therein, OS_NANONSECOND_WAIT() was disabled.
 *
 *          Though their names imply I/O device access, the IO_READ()
 *          and IO_WRITE() macros may also be crafted to support a
 *          memory-mapped controller -- MAC or repeater -- device
 *          interface.  Regardless, they also support the notion of an
 *          <adapter> reference for DRV_OS environments requiring
 *          abstraction to touch hardware, e.g., NDIS, WIN32_DIAG,
 *          NOVELL_CHSM.  Other environments can ignore the <adapter>
 *          or use it as a device base address parameter.
 *--------------------------------------------------------------------*/

#if (DRV_OS == NDIS)            /*------------------------------------*
                                 * [Step 1a] Win95/WinNT NDIS
                                 *
                                 * Assumes <adapter> is the starting
                                 * I/O port address of the MAC device
                                 *------------------------------------*/
#define OS_MICROSECOND_WAIT(u)          NdisStallWrapper(u)
#define IO_WRITE(adapter, port, val)    \
    NdisRawWritePortUlong((PULONG) ((adapter) + (port)), (ULONG) (val))
#define IO_READ(adapter, port)          \
    NdisIoReadWrapper((adapter) + (port))

/***********************************************************************
 * NdisStallWrapper() - "OS-friendly" wait in 100 uSec chunks
 **********************************************************************/
static void
NdisStallWrapper(DWORD num_microseconds)
{
    /*
     * Semi-hog:  relinquish the CPU every 100 microseconds to prevent
     * other drivers (e.g., tty character devices) from starving
     */
    while (num_microseconds > 25) {
        NdisStallExecution(25);
        num_microseconds -= 25;
    }
    NdisStallExecution(num_microseconds);
}

/***********************************************************************
 * NdisIoReadWrapper() - expected NdisRawReadPortUlong() interface
 **********************************************************************/
static DWORD
NdisIoReadWrapper(DWORD port)
{
    ULONG       val;

    NdisRawReadPortUlong((PULONG) port, &val);
    return ((DWORD) val);
}

#elif (DRV_OS == NOVELL_CHSM)   /*------------------------------------*
                                 * [Step 1b] Novell 32-bit ODI
                                 *
                                 * Assumes <adapter> has been cast as
                                 * pointer to a PHY_ADAPTER structure
                                 *------------------------------------*/

#define OS_MICROSECOND_WAIT(u)          Odi32MicroWait(u)
#define OS_NANOSECOND_WAIT(n)           Slow()  /* 0.5 uSec NOP */
#ifndef IO_MAPPED               /* default:  memory-mapped device */
#define IO_WRITE(adapter, port, val)                            \
    Wrt32(((PHY_ADAPTER *) (adapter))->bus_tag, NULL,           \
          (void *) (((port) + (UINT8 *)                         \
                     ((PHY_ADAPTER *) (adapter))->dev_addr)),   \
          (UINT32) (val))
#define IO_READ(adapter, port)          (DWORD)                 \
    Rd32(((PHY_ADAPTER *) (adapter))->bus_tag, NULL,            \
         (void *) (((port) + (UINT8 *)                          \
                   ((PHY_ADAPTER *) (adapter))->dev_addr)))
#else                           /* non-default:  I/O-mapped device */
#define IO_WRITE(adapter, port, val)                            \
    Out32(((PHY_ADAPTER *) (adapter))->bus_tag,                 \
          (void *) (((port) + (UINT8 *)                         \
                     ((PHY_ADAPTER *) (adapter))->dev_addr)),   \
          (UINT32) (val))
#define IO_READ(adapter, port)          (DWORD)                 \
    In32(((PHY_ADAPTER *) (adapter))->bus_tag,                  \
         (void *) (((port) + (UINT8 *)                          \
                   ((PHY_ADAPTER *) (adapter))->dev_addr)))
#endif

/***********************************************************************
 * Odi32MicroWait() - Microsecond delay under 32-bit Novell ODI OS
 **********************************************************************/
static void
Odi32MicroWait(DWORD num_microseconds)
{
    /*
     * CMSMGetMicroTimer() support is suspect under some client
     * operating systems as only 16 bits of returned 32-bit timer value
     * may be valid.  Instead, use the (hopefully) more trustworthy
     * Slow() service.
     */
    while (num_microseconds-- > 0) {
        Slow();         /* 0.5 uSec NOP per CHSM ODI spec v1.11 */
        Slow();
    }
}

#elif (DRV_OS == SCO_MDI)       /*------------------------------------*
                                 * [Step 1c] SCO Unix drivers: MDI,
                                 * LLI, DLPI, and Gemini.
                                 *
                                 * Assumes <adapter> is the starting
                                 * I/O port address of the MAC device
                                 *------------------------------------*/
#define OS_MICROSECOND_WAIT(u)          suspend(u)
#define IO_WRITE(adapter, port, val)    outd((adapter) + (port), (val))
#define IO_READ(adapter, port)          ind((adapter) + (port))

#elif (DRV_OS == SCO_LLI)
#define OS_MICROSECOND_WAIT(u)          TBD
#define IO_WRITE(adapter, port, val)    TBD
#define IO_READ(adapter, port)          TBD

#elif (DRV_OS == SCO_DLPI)
#define OS_MICROSECOND_WAIT(u)          TBD
#define IO_WRITE(adapter, port, val)    TBD
#define IO_READ(adapter, port)          TBD

#elif (DRV_OS == SCO_GEMINI)
#define OS_MICROSECOND_WAIT(u)          TBD
#define IO_WRITE(adapter, port, val)    TBD
#define IO_READ(adapter, port)          TBD

#elif (DRV_OS == SOLARIS)       /*------------------------------------*
                                 * [Step 1d] Sun Solaris driver
                                 *
                                 * Assumes <adapter> is the starting
                                 * I/O port address of the MAC device
                                 *------------------------------------*/
#define OS_MICROSECOND_WAIT(u)          drv_usecwait(u)
#define IO_WRITE(adapter, port, val)    outl((adapter) + (port), (val))
#define IO_READ(adapter, port)          inl((adapter) + (port))

#elif (DRV_OS == LINUX)         /*------------------------------------*
                                 * [Step 1e] Linux driver support
                                 *
                                 * Assumes <adapter> is the starting
                                 * I/O port address of the MAC device
                                 *------------------------------------*/
/*
 * Caution:  per the outb(9) man page, the outl() arguments are the
 * opposite of most DOS-based output routines:  outl(value, portAddr)
 * instead of outl(portAddr, value).
 */
#ifdef __KERNEL__

#include <linux/config.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif


#ifdef VER_CTL
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS			/* force it on */
#endif
#endif


#define OS_MICROSECOND_WAIT(u)          udelay(u)
#else
/* #include <unistd.h> */
/* #define OS_MICROSECOND_WAIT(u)          usleep(u) */
#define OS_MICROSECOND_WAIT(u)          LinuxMicroWait(u)
void LinuxMicroWait(int microseconds)
{
    do {
    } while (--microseconds > 0);
}
#endif
#define IO_WRITE(adapter, port, val)    outl((val), (adapter) + (port))
#define IO_READ(adapter, port)          inl((adapter) + (port))

#elif (DRV_OS == PKTDRV)        /*------------------------------------*
                                 * [Step 1f] DOS/PCTCP packet driver
                                 *
                                 * Assumes <adapter> is the starting
                                 * I/O port address of the MAC device
                                 *------------------------------------*/
#include <???.h>

#define OS_MICROSECOND_WAIT(u)          TBD
#define IO_WRITE(adapter, port, val)    TBD
#define IO_READ(adapter, port)          TBD

#elif (DRV_OS == WIN32_DIAG)    /*------------------------------------*
                                 * [Step 1g] RSS' Win95/WinNT Diagnostic
                                 *
                                 * Assumes <adapter> is the adapter
                                 * handle used by the diagnostic.  Also,
                                 * nullifies OS_NANOSECOND_WAIT(), as
                                 * IO_READ() and IO_WRITE() driver
                                 * ioctl's provide sufficient clock
                                 * spacing, i.e., >> one uSec.
                                 *------------------------------------*/

#define Unknown_OS      0
#define Windows95_OS    1
#define Windows98_OS    2
#define WindowsNT4_OS   3

#define OS_MILLISECOND_WAIT(m)                  SleepEx((m), FALSE)
#define OS_NANOSECOND_WAIT(n)                   /* No delay */

#define IO_WRITE(adapter, port, val)            \
    (void) _PutCSRRegisterWrapper((ULONG) adapter, (ULONG) (port), (ULONG) (val))

#define IO_READ(adapter, port)          \
    _GetCSRRegisterWrapper((ULONG) (adapter), (ULONG) (port))

/***********************************************************************

 * _PutCSRRegisterWrapper() - expected _PutCSRRegister() interface
 **********************************************************************/
static DWORD
_PutCSRRegisterWrapper(ULONG adapter, ULONG port, ULONG val)
{

    HANDLE      drvhandle;
    DWORD       RunningOS;

    RunningOS = CheckOS();

    if (RunningOS != WindowsNT4_OS) {
        _getdriverhandle(&drvhandle);   /* get handle to driver */
        
        if (!drvhandle)                 /* if fail, return -1 */
            return (-1);
    }

    (void) _putcsrregister(drvhandle,adapter, port, val);

    return ((DWORD) val);
}

/***********************************************************************
 * _GetCSRRegisterWrapper() - expected _GetCSRRegister() interface
 **********************************************************************/
static DWORD
_GetCSRRegisterWrapper(ULONG adapter, ULONG port)
{
    ULONG       val = -1;
    HANDLE      drvhandle;
    DWORD       RunningOS;

    RunningOS = CheckOS();

    if (RunningOS != WindowsNT4_OS) {
        _getdriverhandle(&drvhandle);   /* get handle to driver */
        
        if (!drvhandle)                 /* if fail, return -1 */
            return (-1);
    }
    
    (void) _getcsrregister(drvhandle,adapter, port, &val);

    return ((DWORD) val);
}

#else                           /*------------------------------------*
                                 * [Step 1h] Miscellaneous (DOS, etc.)
                                 *
                                 * Assumes <adapter> is the starting
                                 * I/O port address of the MAC device
                                 *------------------------------------*/
#include <conio.h>

#if   defined(__WATCOMC__)
#elif defined(__BORLANDC__)
#elif defined(_MSVC_VER)
#endif

#ifdef _MAPTEST_
BYTE    phy_data[0x10] = {0};   /* DATA - initialized */
BYTE    phy_bss[0x10];          /* BSS  - uninitialized */
#endif

extern DWORD    inpd(int port);
extern DWORD    outpd(int port, DWORD val);
#define OS_MICROSECOND_WAIT(u)          DosMicroWait(u)
#define IO_WRITE(adapter, port, val)    \
    (void) outpd((int) (adapter) + (port), (val))
#define IO_READ(adapter, port)          \
    inpd((int) (adapter) + (port))

/*
 * Convert microseconds to PC system ticks.  The 16-bit Timer0, in
 * Mode 3 (square wave), decrements two counts per 1.1932 MHz input
 * clock on most systems.  We round this 2.3864 ticks/uSec to integer
 * (239 / 100) here, yielding a 32-bit unsigned range of 17.971 seconds.
 * Warning:  Some AMI BIOSes have a bug, configuring Timer0 into Mode 2
 * (rate generator), which only decrements one count per clock.  On
 * such systems, we'll suffer a delay that is twice as long as
 * expected.  This could be a problem for longer, millisecond-to-second
 * duration delays, however, most MII related delays are microsecond
 * or less.
 */
#define PC_TIMER0_TICKS(usec)           \
    (1 + (((DWORD) (usec) * 239) / 100))

/***********************************************************************
 * DosMicroWait() - Microsecond delay under 16-bit DOS environment
 **********************************************************************/
static void
DosMicroWait(DWORD count)
{
    DWORD       t0;
    DWORD       t1;
    DWORD       delta_t;
    BYTE        timer0_lsb;

    /*
     * Writing a command value of 0x00 to Port 0x43 transfers
     * the Timer 0 counter into the output latch register.  The value
     * is held until read from port 0x40 (two successive byte reads,
     * LSB followed by the MSB).  A truly bullet-proof implementation
     * would lock interrupts prior to latching and reading the counter.
     */
    outp(0x43, 0x00);
    timer0_lsb = inp(0x40);
    t1 = ((DWORD) inp(0x40) << 8) + timer0_lsb;
    delta_t = 0;
    count = PC_TIMER0_TICKS(count);     /* uSec to Timer0 ticks */

    do {
        count -= delta_t;
        t0 = t1;
        outp(0x43, 0x00);
        timer0_lsb = inp(0x40);
        t1 = ((DWORD) inp(0x40) << 8) + timer0_lsb;
        if (t1 > t0) {
            t0 += 0x10000UL;            /* handle 16-bit rollover */
        }
        delta_t = t0 - t1;
    } while (count > delta_t);
}

#ifdef UNSUPPORTED      /* Most C/C++ only support 16-bit in-line ASM */
DWORD
outpd(int port, DWORD val)
{
    __asm {
        mov     dx, port;
        mov     eax, val;
        out     dx, eax;
    }
    return (val);
}

DWORD
inpd(int port)
{
    DWORD       val;

    __asm {
        mov     dx, port;
        in      eax, dx;
        mov     val, eax;
    }
    return (val);
}
#endif  /* UNSUPPORTED, as most C/C++ only support 16-bit in-line ASM */

#endif                          /*------------------------------------*
                                 * [Step 1] End of first porting step
                                 *------------------------------------*/

/*
 * Post- [Step 1] timer derivation.  OS_MICROSECOND_WAIT() and/or
 * OS_MILLISECOND_WAIT() must be #defined for this to work....
 */

#ifndef OS_MICROSECOND_WAIT
#define OS_MICROSECOND_WAIT(u)                  \
    OS_MILLISECOND_WAIT((DWORD) ((u) + 999) / 1000)
#endif

#ifndef OS_MILLISECOND_WAIT
#define OS_MILLISECOND_WAIT(m)  MICROSECOND_WAIT((DWORD) (m) * 1000)
#endif

#ifndef OS_NANOSECOND_WAIT
#define OS_NANOSECOND_WAIT(n)   \
    MICROSECOND_WAIT((DWORD) ((n) + 999) / 1000)
#endif
#define NANOSECOND_WAIT         OS_NANOSECOND_WAIT

/*--------------------------------------------------------------------*
 * [Step 2] Enter (NIC or repeater) adapter's MAC controller-specific
 *          register information.  The MII (MDIO) and any optional
 *          adapter-unique general purpose I/O (GPIO) settings must be
 *          captured here.  In particular, the gpio[] table should be
 *          modified per adapter configuration of the MAC's GPIO
 *          interface [Step 2a].  If additional, custom configuration
 *          is required, the low-level MiiPhyHwReset() routine may be
 *          modifified [Step 2b].
 *
 *          Warning:  at present, only RSS and DEC-compatible fast
 *          ethernet MACs are supported.  Common repeater interfaces
 *          will be added as development (COMET?) progresses at RSS.
 *--------------------------------------------------------------------*/

/* RSS 1161x MACs' CSR9 (MDIO) register offset and values */
#define MAC_MDIO_REG            0x48
#define MDIO_RD_DATA            0x80000L
#define MDIO_RD_ENABLE          0x40000L
#define MDIO_WR_ENABLE          0x00000L
#define MDIO_WR_DATA_1          0x20000L        
#define MDIO_WR_DATA_0          0x00000L        
#define MDIO_CLK                0x10000L
#define MDIO_MASK               0xF0000L
#define SPR_FRCS                0x00000010L /* 11623 serial Flash ROM Chip Select */

/*
 * [Step 2a] MiiPhyHwReset() only affects the individual GPIO bit
 * passed via the PHY_HWOPTS_RESET_GP* argument.  Other GPIO bits are
 * configured as input and written (output) a don't-care zero data
 * value unless the GPIO_DEFAULT_CTRL and GPIO_DEFAULT_DATA values
 * are #defined to contain one or more of the R15_GP[3:0] bit settings.
 * If #defined, GPIO_DEFAULT_[CTRL|DATA] are translated from R15
 * register format (bits 19:16) to R12 format (bits 3:0).  Note that
 * usage of the R12 GPIO register was obsoleted upon introduction of the
 * RSS 11617 (a.k.a. the RSS 11611 v5) production device.  Therein, R15
 * is used to control GPIO.  R12 support is provided for legacy support
 * of pre-production 11611 (v3 and v4), 11610, and 11601 devices.
 *
 * If GPIO_DEFAULT_CTRL is #undefined, then we'll default to
 * RSS11623_GPIO_CTRL (GP1 and GP2 output) plus the RSS reference
 * design convention of GP0 for a PHY hardware reset.  If
 * GPIO_DEFAULT_CTRL is #defined as zero then only the
 * PHY_HWOPTS_HWRESET_GP[0-3]* <hw_options> switch is
 * configured as an output signal; the remaining GPIO lines are
 * configured as input signals.  The default RSS11623 value is
 * highly recommended.
 *
 * If GPIO_DEFAULT_CTRL is non-zero, then in addition to the PHY HW
 * reset GPIO line, any of the other GP[3:0] signals contained in
 * in that GPIO_DEFAULT_CTRL bit-mask are also configured for output.
 * The values written to these additional output GPIO signals depend
 * on the state of the conditional GPIO_DEFAULT_DATA macro.  If
 * GPIO_DEFAULT_DATA is #undefined, then the values written out are
 * based on values which are read from the GPIO register.  This allows
 * retention of GPIO outputs previously configured by another software
 * driver function.  If GPIO_DEFAULT_DATA is #defined, then the other
 * GPIO values will be forced to the GPIO_DEFAULT_DATA value.
 *
 * Finally, GPIO_INTERWORD_GAP, the microsecond-granularity R12 or R15
 * inter-word write time delay, and GPIO_SETTLE_MICROSEC, the
 * millisecond-granularity post-reset hardware settle time constants can
 * be overridden via conditional compilation special PHY hardware
 * circumstances.
 */
#define MAC_R15_GPIO_REG        0x78            /* MAC >= RSS 11617 */
#define R15_GPIO_CTRL           0x08000000L     /* bit-27 */
#define R15_GP0                 0x00010000L     /* bit-16 */
#define R15_GP1                 0x00020000L     /* bit-17 */
#define R15_GP2                 0x00040000L     /* bit-18 */
#define R15_GP3                 0x00080000L     /* bit-19 */
#define R15_GP0_THRU_GP3        (R15_GP0 | R15_GP1 | R15_GP2 | R15_GP3)
#define R15_GPIO_MASK           (R15_GPIO_CTRL | R15_GP0_THRU_GP3)
#define R15_WRITE_ONLY_BITS     0x0FF00000L     /* read as ones on RSS-
                                                 * compatible devices */

#define MAC_R12_GPIO_REG        0x60            /* MAC <= RSS 11611 */
#define R12_GPIO_CTRL           0x00000100L     /* bit-8 */
#define R12_GP0                 0x00000001L     /* bit-0 */
#define R12_GP1                 0x00000002L     /* bit-1 */
#define R12_GP2                 0x00000004L     /* bit-2 */
#define R12_GP3                 0x00000008L     /* bit-3 */
#define R12_GP0_THRU_GP3        (R12_GP0 | R12_GP1 | R12_GP2 | R12_GP3)
#define R12_GPIO_MASK           (R12_GPIO_CTRL | R12_GP0_THRU_GP3)

/*
 * RSS 11623 reserves GP[3:1] for EEPROM/SPI access as follows
 *
 * GP0 - input/available (typically used for MII PHY hardware reset)
 * GP1 - output EECK
 * GP2 - output EEDI [sic]
 * GP3 - input  EEDO [sic]
 */
#define RSS11623_GPIO_CTRL      (R15_GP1 | R15_GP2)     /* outputs */

#ifndef GPIO_DEFAULT_CTRL
#define R15_GPIO_DEFAULT_CTRL   (R15_GP0 | RSS11623_GPIO_CTRL)  // was 0L
#define R15_GPIO_DEFAULT_DATA   0L

#else
#define R15_GPIO_DEFAULT_CTRL   (GPIO_DEFAULT_CTRL & R15_GP0_THRU_GP3)

#ifndef GPIO_DEFAULT_DATA
#define R15_GPIO_DEFAULT_DATA   0L
#else
#define R15_GPIO_DEFAULT_DATA   (GPIO_DEFAULT_DATA & R15_GP0_THRU_GP3)
#endif  /* GPIO_DEFAULT_DATA */

#endif  /* GPIO_DEFAULT_CTRL */

#define R12_GPIO_DEFAULT_CTRL   ((DWORD) R15_GPIO_DEFAULT_CTRL >> 16)
#define R12_GPIO_DEFAULT_DATA   ((DWORD) R15_GPIO_DEFAULT_DATA >> 16)

#ifdef TEST_R12                 /* ################################## */
/* Coerce R15 as R12 for testing with pre-11617 devices */
#undef  MAC_R15_GPIO_REG
#undef  R15_GPIO_CTRL
#undef  R15_GP0
#undef  R15_GP1
#undef  R15_GP2
#undef  R15_GP3
#undef  R15_GP0_THRU_GP3
#undef  R15_GPIO_MASK
#undef  R15_WRITE_ONLY_BITS

#define MAC_R15_GPIO_REG        MAC_R12_GPIO_REG
#define R15_GPIO_CTRL           R12_GPIO_CTRL
#define R15_GP0                 R12_GP0
#define R15_GP1                 R12_GP1
#define R15_GP2                 R12_GP2
#define R15_GP3                 R12_GP3
#define R15_GP0_THRU_GP3        R12_GP0_THRU_GP3
#define R15_GPIO_MASK           R12_GPIO_MASK
#define R15_WRITE_ONLY_BITS     0x00000000L

#undef  R12_GPIO_DEFAULT_CTRL
#undef  R12_GPIO_DEFAULT_DATA
#undef  R15_GPIO_DEFAULT_CTRL
#undef  R15_GPIO_DEFAULT_DATA

#ifndef GPIO_DEFAULT_CTRL
#define R15_GPIO_DEFAULT_CTRL   0L
#define R15_GPIO_DEFAULT_DATA   0L
#else
#define R15_GPIO_DEFAULT_CTRL   (GPIO_DEFAULT_CTRL & R15_GP0_THRU_GP3)
#ifndef GPIO_DEFAULT_DATA
#define R15_GPIO_DEFAULT_DATA   0L
#else
#define R15_GPIO_DEFAULT_DATA   (GPIO_DEFAULT_DATA & R15_GP0_THRU_GP3)
#endif  /* GPIO_DEFAULT_DATA */
#endif  /* GPIO_DEFAULT_CTRL */
#define R12_GPIO_DEFAULT_CTRL   R15_GPIO_DEFAULT_CTRL   /* was >> 16 */
#define R12_GPIO_DEFAULT_DATA   R15_GPIO_DEFAULT_DATA   /* was >> 16 */

#endif                          /* ################################## */

/* R12 or R15 inter-word register write timing */
#ifndef GPIO_INTERWORD_GAP
#define GPIO_INTERWORD_GAP      1       /* microseconds */
#endif

/* Post-HW reset settle time */
#ifndef GPIO_SETTLE_TIME
#define GPIO_SETTLE_TIME        1       /* milliseconds */
#endif

typedef struct _gpio_ {
    DWORD       io_ctrl;
    DWORD       io_data;
} GPIO;

static GPIO     r12_gpio = {
    R12_GPIO_DEFAULT_CTRL | R12_GPIO_CTRL,
    R12_GPIO_DEFAULT_DATA
};

static GPIO     r15_gpio = {
    R15_GPIO_DEFAULT_CTRL | R15_GPIO_CTRL,
    R15_GPIO_DEFAULT_DATA
};


/* MII frame fields */
#define MII_READ_ST_OP          0x6     /* MII READ start, opcode */
#define MII_WRITE_ST_OP         0x5     /* MII WRITE start, opcode */
#define MII_WRITE_TA            0x2     /* MII WRITE turnaround bits */
#define MII_WRITE_FRAME         \
    (((DWORD) MII_WRITE_ST_OP << 28) | ((DWORD) MII_WRITE_TA << 16))

/* Local (static) function prototypes */
static void     MiiPreamble(PHY *phy);
static void     MiiIdle(PHY *phy, int num_clocks);

/***********************************************************************
 * MiiRead() - read serial data from MII register
 **********************************************************************/
WORD
MiiRead(PHY *phy, BYTE reg_addr)
{
    DWORD       io_data;
    WORD        mii_data;
    int         n;

    /* Validate the device post-PhyInit() */
    if (IS_PHY_INITIALIZED(phy) && !(phy->devices & PHY_PORT_MII)) {
        return (0x0000);
    }

    /* Output 32-bit preamble (all ones) to re-sync the PHY */
    MiiPreamble(phy);

    /* Preserve non-MDIO bits in shared MAC_MDIO_REG (SPR) */
    io_data = IO_READ(phy->adapter, MAC_MDIO_REG) & ~MDIO_MASK;

    /* Write the first 14 <mii_data> frame bits in MSB-first order */
    mii_data =
        (MII_READ_ST_OP << 10) | (phy->addr << 5) | (reg_addr & 0x1F);
    for (n = 0x2000; n > 0; n >>= 1) {
        io_data &= ~MDIO_MASK;          /* Preserve non-MDIO bits */
        if (mii_data & n) {
            io_data |= (MDIO_WR_ENABLE | MDIO_WR_DATA_1);
        } else {
            io_data |= (MDIO_WR_ENABLE | MDIO_WR_DATA_0);
        }
        IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data);
        NANOSECOND_WAIT(200);
        IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data | MDIO_CLK);
        NANOSECOND_WAIT(200);
    }

    /*
     * Read the next 18 bits [17:0] in MSB-first order.  The first two
     * bits, turnaround (TA) high-Z and zero, are effectively ignored.
     */
    mii_data = 0;
    io_data &= ~MDIO_MASK;          /* Preserve non-MDIO bits */
    io_data |= MDIO_RD_ENABLE;
    for (n = 0; n < 18; n++) {
        IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data);
        NANOSECOND_WAIT(200);
        /*
         * ??? Shouldn't IO_READ() occur *after* MDC_CLK rising edge?
         * The IEEE 802.3u specification -- Clause 22.3.4, "MDIO timing
         * relationship to MDC" and Figure 22-17, "MDIO sourced by
         * PHY" -- doesn't imply that data is latched on the falling
         * edge of MDC.  In any case, that is what works with a National
         * Semiconductor DP83840 PHY, but may be a source of problems
         * with other PHY devices....
         */
        mii_data <<= 1;
        if (IO_READ(phy->adapter, MAC_MDIO_REG) & MDIO_RD_DATA) {
            mii_data |= 0x0001;
        }
        IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data | MDIO_CLK);
        NANOSECOND_WAIT(300);        /* [sic] 300 nSec PHY hold */
    }

    /* Output a single, idle (high-Z, tri-stated) clock transition */
    MiiIdle(phy, 1);

    return (mii_data & 0xFFFF);
}

/***********************************************************************
 * MiiWrite() - write serial data to MII register
 **********************************************************************/
void
MiiWrite(PHY *phy, BYTE reg_addr, WORD reg_value)
{
    DWORD       io_data;
    DWORD       mii_frame;
    DWORD       mask;

    /* Suppress in read-only (debug) mode or when device not present */
    if (phy->read_only || !(phy->devices & PHY_PORT_MII)) {
        return;

    }

    /* Output 32-bit preamble (ones) sequence to re-sync the PHY */
    MiiPreamble(phy);

    /* Preserve non-MDIO bits in shared MAC_MDIO_REG (SPR) */
    io_data = IO_READ(phy->adapter, MAC_MDIO_REG) & ~MDIO_MASK;

    /* Form and write 32-bit [31:0] <mii_frame> in MSB-first order */
    mii_frame = MII_WRITE_FRAME | ((DWORD) phy->addr << 23) |
        ((DWORD) (reg_addr & 0x1F) << 18) | (DWORD) reg_value;
    for (mask = 0x80000000UL; mask > 0; mask >>= 1) {
        io_data &= ~MDIO_MASK;          /* Preserve non-MDIO bits */
        if (mii_frame & mask) {
            io_data |= (MDIO_WR_ENABLE | MDIO_WR_DATA_1);
        } else {
            io_data |= (MDIO_WR_ENABLE | MDIO_WR_DATA_0);
        }
        IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data);
        NANOSECOND_WAIT(200);
        IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data | MDIO_CLK);
        NANOSECOND_WAIT(200);
    }

    /* Output a single, idle (high-Z, tri-stated) clock transition */
    MiiIdle(phy, 1);
}

/***********************************************************************
 * MiiPreamble() - Generate 32-bit MII preamble sequence (all ones)
 **********************************************************************/
static void
MiiPreamble(PHY *phy)
{
    DWORD       io_data;
    int         n;

    /* Preserve non-MDIO bits in shared MAC_MDIO_REG (SPR) */
    io_data = IO_READ(phy->adapter, MAC_MDIO_REG) & ~MDIO_MASK;

    if (!phy->preamble_suppression) {
        io_data |= (MDIO_WR_ENABLE | MDIO_WR_DATA_1);
        for (n = 0; n < 32; n++) {
            IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data);
            NANOSECOND_WAIT(200);
            IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data | MDIO_CLK);
            NANOSECOND_WAIT(200);
        }
    }
}

/***********************************************************************
 * MiiIdle() - Generate one or more bits of MII idle sequence (high-Z)
 **********************************************************************/
static void
MiiIdle(PHY *phy, int num_clocks)
{
    DWORD       io_data;

    /* Preserve non-MDIO bits in shared MAC_MDIO_REG (SPR) */
    io_data = IO_READ(phy->adapter, MAC_MDIO_REG) & ~MDIO_MASK;
    io_data |= MDIO_RD_ENABLE;

    do {
        IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data);
        NANOSECOND_WAIT(200);
        IO_WRITE(phy->adapter, MAC_MDIO_REG, io_data | MDIO_CLK);
        NANOSECOND_WAIT(200);
    } while (--num_clocks > 0);
}

/***********************************************************************
 * MiiMillisecondWait() - OS-dependent busy-wait/process sleep wrapper
 **********************************************************************/
void
MiiMillisecondWait(DWORD num_milliseconds)
{
    OS_MILLISECOND_WAIT(num_milliseconds);
}

/***********************************************************************
 * MiiMicrosecondWait() - OS-dependent busy-wait/process sleep wrapper
 **********************************************************************/
void
MiiMicrosecondWait(DWORD num_microseconds)
{
    OS_MICROSECOND_WAIT(num_microseconds);
}

/***********************************************************************
 * MiiPhyHwReset() - Board-specific hardware RESET
 **********************************************************************/
void
MiiPhyHwReset(PHY *phy, DWORD hw_options)
{
    /*
     * [Step 2b] Provide customized, board-specific initialization
     *           here as necessary.  Hopefully, the fifteen
     *           PHY_HWOPTS_HWRESET_GP* options, in conjunction with
     *           the GPIO_DEFAULT_CTRL, GPIO_DEFAULT_DATA,
     *           GPIO_INTERWORD_GAP and GPIO_SETTLE_TIME conditional
     *           compilation switches will provide enough flexibility
     *           to avoid customized code changes herein.
     */

    WORD        gpio_reg = MAC_R15_GPIO_REG;
    BOOLEAN     gpio_not = FALSE;
    GPIO        gpio;
    DWORD       gpio_preserved;
    DWORD       gpio_bits;

    if (hw_options & PHY_HWOPTS_GP0_SPI_DOUT) {
        DWORD       spr_value = SPR_FRCS | IO_READ(phy->adapter, MAC_MDIO_REG);

        /* Set GPIO[0] to input. It will be used for SPI_DOUT */
        gpio = r15_gpio;
        IO_WRITE(phy->adapter, gpio_reg, gpio.io_ctrl &~0x10000);
        /* Use ROM_CS# for CN7221 (11625) Ramses RESET# */
        IO_WRITE(phy->adapter, MAC_MDIO_REG, spr_value);
        MICROSECOND_WAIT(GPIO_INTERWORD_GAP);
        IO_WRITE(phy->adapter, MAC_MDIO_REG, (spr_value & ~SPR_FRCS));
        MILLISECOND_WAIT(GPIO_SETTLE_TIME);
		/* Reset GPC bit */
	    gpio_preserved = IO_READ(phy->adapter, gpio_reg);
        gpio_preserved &= ~R15_WRITE_ONLY_BITS;
		IO_WRITE(phy->adapter, gpio_reg, gpio_preserved);

        return;
    }
	
    switch (hw_options &= PHY_HWOPTS_HWRESET_MASK) {
        /* Active-high, R15.GP[3:0] */
        default:
        case PHY_HWOPTS_HWRESET_GP0:
            gpio_bits = R15_GP0; break;
        case PHY_HWOPTS_HWRESET_GP1:
            gpio_bits = R15_GP1; break;
        case PHY_HWOPTS_HWRESET_GP2:
            gpio_bits = R15_GP2; break;
        case PHY_HWOPTS_HWRESET_GP3:
            gpio_bits = R15_GP3; break;

        /* Active-low <gpio_not> R15.GP[3:0] */
        case PHY_HWOPTS_HWRESET_GP0_NOT:
            gpio_bits = R15_GP0; gpio_not = TRUE; break;
        case PHY_HWOPTS_HWRESET_GP1_NOT:
            gpio_bits = R15_GP1; gpio_not = TRUE; break;
        case PHY_HWOPTS_HWRESET_GP2_NOT:
            gpio_bits = R15_GP2; gpio_not = TRUE; break;
        case PHY_HWOPTS_HWRESET_GP3_NOT:
            gpio_bits = R15_GP3; gpio_not = TRUE; break;

        /* Pre-production RSS 11611 active-high R12.GP[3:0] */
        case PHY_HWOPTS_HWRESET_GP0_R12:
            gpio_bits = R12_GP0; gpio_reg = MAC_R12_GPIO_REG; break;
        case PHY_HWOPTS_HWRESET_GP1_R12:
            gpio_bits = R12_GP1; gpio_reg = MAC_R12_GPIO_REG; break;
        case PHY_HWOPTS_HWRESET_GP2_R12:
            gpio_bits = R12_GP2; gpio_reg = MAC_R12_GPIO_REG; break;
        case PHY_HWOPTS_HWRESET_GP3_R12:
            gpio_bits = R12_GP3; gpio_reg = MAC_R12_GPIO_REG; break;

        /* Wildcard _ALL GP[3:0] registers */
        case PHY_HWOPTS_HWRESET_ALL:
            gpio_bits = R15_GP0_THRU_GP3; break;
        case PHY_HWOPTS_HWRESET_ALL_NOT:
            gpio_bits = R15_GP0_THRU_GP3; gpio_not = TRUE; break;
        case PHY_HWOPTS_HWRESET_ALL_R12:
            gpio_bits = R12_GP0_THRU_GP3;
            gpio_reg = MAC_R12_GPIO_REG; break;
    }

    /*
     * Write the I/O control word, which configures GP[3:0] lines as
     * output or input signals.  The 1161x has an R12 sticky-bit
     * bug requiring a second, clearing write of the R12_GPIO_CTRL bit.
     * Also, since the GPIO registers may have bits used by other
     * functions, the <gpio_preserved> register read ensures that
     * these values are unaffected by our GPIO write logic.
     */
    gpio_preserved = IO_READ(phy->adapter, gpio_reg);
    if (gpio_reg == MAC_R12_GPIO_REG) {
        /*
         * Obsolete R12 register support for earlier, pre-production
         * RSS 11606, 11610, 11611 (v3 and v4) and compatible devices.
         */
        phy->r12_gpio = gpio_reg;
        gpio = r12_gpio;
        gpio.io_ctrl |= (gpio_preserved & ~R12_GPIO_MASK);
#ifdef GPIO_DEFAULT_DATA
        /* Force GPIO_DEFAULT_DATA output values */
        gpio.io_data |= (gpio_preserved & ~R12_GPIO_MASK);
#else
        /* Preserve current GPIO register output values */
        gpio.io_data |= (gpio_preserved & ~R12_GPIO_CTRL);
#endif
        /* Additional R12_GPIO_CTRL sticky-bit set followed by clear */
        IO_WRITE(phy->adapter, gpio_reg, gpio.io_ctrl | gpio_bits);
        gpio.io_ctrl &= ~R12_GPIO_CTRL;
        // MICROSECOND_WAIT(GPIO_INTERWORD_GAP);

    } else {
        /*
         * Nominal, GPIO register R15 usage.  Be certain to clear write-
         * only bits (reserved on the RSS 11617; always read as one
         * on RSS-compatible devices) as these may interfere with
         * proper GPIO port selection and assignment on RSS-compatible
         * MAC devices.
         */
        gpio_preserved &= ~R15_WRITE_ONLY_BITS;
        gpio = r15_gpio;
        gpio.io_ctrl |= (gpio_preserved & ~R15_GPIO_MASK);
#ifdef GPIO_DEFAULT_DATA
        /* Force GPIO_DEFAULT_DATA output values */
        gpio.io_data |= (gpio_preserved & ~R15_GPIO_MASK);
#else
        /* Preserve current GPIO register output values */
        gpio.io_data |= (gpio_preserved & ~R15_GPIO_CTRL);
#endif
    }
    IO_WRITE(phy->adapter, gpio_reg, gpio.io_ctrl | gpio_bits);
    // MICROSECOND_WAIT(GPIO_INTERWORD_GAP);

    /*
     * Toggle the I/O data word's <gpio_bits> from zero to one and back
     * to zero, unless inverted -- i.e., <gpio_not> -- which goes from 
     * one to zero and back to one.  The GPIO_INTERWORD_GAP effects
     * the positive or negative (not) pulse duration.
     */
    if (gpio_not) {
        gpio.io_data &= ~gpio_bits;     /* clears <gpio_bits> */
    } else {
        gpio.io_data |= gpio_bits;      /* sets   <gpio_bits> */
    }
    IO_WRITE(phy->adapter, gpio_reg, gpio.io_data);
    MICROSECOND_WAIT(GPIO_INTERWORD_GAP);
    IO_WRITE(phy->adapter, gpio_reg, (gpio.io_data ^ gpio_bits));

    /*
     * Allow for post-HW reset settle time which may be required before
     * manipulating the device via its MII serial management registers.
     */
    MILLISECOND_WAIT(GPIO_SETTLE_TIME);
}

/***********************************************************************
 * MiiIoRead() - Export wrapper for the IO_READ() macro
 **********************************************************************/
DWORD
MiiIoRead(PHY  *phy, WORD io_offset)
{
    return (IO_READ(phy->adapter, io_offset));
}

/***********************************************************************
 * MiiIoRead() - Export wrapper for the IO_WRITE() macro
 **********************************************************************/
void
MiiIoWrite(PHY *phy, WORD io_offset, DWORD io_value)
{
    IO_WRITE(phy->adapter, io_offset, io_value);
}

#if (DRV_OS == DOS)     /* ########################################## */

/***********************************************************************
 * MiiSetBits() - Set (logical-OR) one or more bits in MII register
 **********************************************************************/
void
MiiSetBits(PHY *phy, BYTE reg_addr, WORD bits)
{
    MiiWrite(phy, reg_addr, MiiRead(phy, reg_addr) | bits);
}

/***********************************************************************
 * MiiClrBits() - Clear (logical-NAND) one or more bits in MII register
 **********************************************************************/
void
MiiClrBits(PHY *phy, BYTE reg_addr, WORD bits)
{
    MiiWrite(phy, reg_addr, MiiRead(phy, reg_addr) & ~bits);
}

#endif                  /* ########################################## */
