/*************************************************************************************************
 *                         PVCS Version Control Information                                      *
 *                                                                                               *
 * $Header:   P:\d878\HomeLAN\Archives\NdisDriver\MII.H_v   1.1   May 12 1999 12:05:54   wangmx  $                            
 *                                                                                               *
 ************************************************************************************************/

/*
 *  Module Name: mii.h
 *  Author: Joe Bonaker
 *
 *  Copyright  1997-1998 Conexant Systems, Inc.
 *  All rights reserved.
 *
 *  This software may be used and distributed according to the terms
 *  of the GNU Public License, incorporated herein
 */

/*
 *  Module Description: The 100Base-T serial Media Independent Interface
 *  (MII), which defines OS- and adapter-dependent functions imported by
 *  the PHY driver.  The accompanying "mii.c" module must be tailored to
 *  adhere to the implementation-dependent environment.
 *
 * $Log:   P:\d878\HomeLAN\Archives\NdisDriver\MII.H_v  $
 * 
 *    Rev 1.1   May 12 1999 12:05:54   wangmx
 * Various driver code enhancements are done
 * to the base line V1.01.
 * Revision 1.8  1998/09/25 16:10:30  bonakeaj
 * - Integrated 1M8 HomePNA with multi-PHY support
 * Revision 1.7  1998/02/11 23:26:06  bonakeaj
 * - RSS 11617 R15 GPIO PHY_HWRESET(), WIN32_DIAG NT 4.0 merge
 * Revision 1.6  1997/09/12 00:10:04  bonakeaj
 * - DRV_OS=LINUX port
 * Revision 1.5  1997/08/21 21:55:00  kamstrd
 * - Updated copyright headers
 * - Replaced RW116?? with RSS116??
 * - Replaced RW11606 with RSS1161x in some comments
 * Revision 1.4  1997/07/31 13:45:52  bonakeaj
 * - DRV_OS=NOVELL_CHSM changes (minor) to support Novell
 * - CHSM-32 driver integration
 * Revision 1.3  1997/07/21 14:35:26  bonakeaj
 * - Cast MII_SET/CLR_MASK() macros to eliminate MSVC warnings
 * Revision 1.2  1997/07/12 02:24:18  bonakeaj
 * - Added PHY_LINK_NEW_MEDIA support, minor type changes for
 * - NDIS 3.1 macApi/phyDrv integration, and fixed PhyInit() bug where
 * - PhyDisable()'s isolate action prevented Auto-Negotiation
 * - completion with the NSC DP83840 PHY device.
 **********************************************************************/

#ifndef _MII_H_
#define _MII_H_

#include "phytypes.h"

/*
 * Special "adapter" handle for 32-bit Novell ODI drivers.  Most
 * other drivers assume that the phy->adapter value is simply the
 * starting I/O port (or memory-mapped) address of the MAC (or repeater)
 * controller device.
 */
typedef struct {
    void       *dev_addr;       /* MAC's I/O or memory base address */
    void       *bus_tag;        /* CMSMSearchAdapter() busTag value */
} PHY_ADAPTER;

/*
 * Optional PHY hardware reset support; see the "mii.c" module for
 * adapter-specific implementation details.  If the adapter does not
 * support PHY hardware reset (e.g., an RSS1161x MAC GPIO pin tied to
 * the PHY's RESET ping), one can simply #define an alternate, null
 * PHY_HWRESET macro prior to #including this file:
 *
 *     ...
 *     #define PHY_HWRESET(phy, hw_opts)
 *     #include "mii.h"
 *     ...
 */
#ifndef PHY_HWRESET
extern  void                    MiiPhyHwReset(PHY  *phy,
                                              DWORD hw_opts);
#define PHY_HWRESET             MiiPhyHwReset
#endif

/*
 * MII_SET_BIT, MII_SET_MASK(), MII_CLR_BIT(), MII_CLR_MASK()
 *
 * Macros which set or clear one or more MII register bits.  In DOS
 * mode, a few hundred bytes can be saved by calling MiiSetBits()
 * and MiiClearBits() instead of the in-line MiiRead() and MiiWrite()
 * calls, as these macros are used extensively.
 */
#if (DRV_OS == DOS)     /* ########################################## */

extern void             MiiSetBits(PHY *phy, BYTE reg_addr, WORD bits);
extern void             MiiClrBits(PHY *phy, BYTE reg_addr, WORD bits);
#define MII_SET_MASK    MiiSetBits
#define MII_CLR_MASK    MiiClrBits

#else                   /* ########################################## */

#define MII_SET_MASK(phy, regaddr, mask)        \
    MiiWrite((phy),                             \
             (BYTE) (regaddr),                  \
             (WORD) (MiiRead((phy), (BYTE) (regaddr)) | (mask)))
#define MII_CLR_MASK(phy, regaddr, mask)        \
    MiiWrite((phy),                             \
             (BYTE) (regaddr),                  \
             (WORD) (MiiRead((phy), (BYTE) (regaddr)) & ~(mask)))
#endif                  /* ########################################## */

#define MII_SET_BIT     MII_SET_MASK
#define MII_CLR_BIT     MII_CLR_MASK

/*
 * MiiRead(), MiiWrite()
 *
 * (MAC or repeater) User-supplied MII read and write functions.  The
 * PHY's serial MII registers cannot be directly addressed via the
 * (PCI) bus, as the PHY does not implement a 16-bit parallel register
 * interface.  Instead, the PHY's serial Media Independent Interface
 * (MII) must be accessed and accumulated into an external, general
 * general purpose port, e.g., an MII port on a LAN adapter MAC or
 * or a repeater controller device.  See the "mii.c" porting file for
 * implementation-dependent MiiRead() and MiiWrite() example code.
 */
extern WORD     MiiRead(PHY *phy, BYTE reg_addr);
extern void     MiiWrite(PHY *phy, BYTE reg_addr, WORD reg_value);

/*
 * MILLISECOND_WAIT() and MICROSECOND_WAIT() - hook for implementation
 * and OS-dependent delay routines defined by the user.  See "mii.c"
 * MiiMillisecondWait() and MiiMicrosecondWait() functions for
 * OS-specific example code.  Two functions are declared to allow the
 * designer to deploy different wait mechanisms.  Short, microsecond
 * delays may be implemented via a non-blocking, busy-wait mechanism
 * whereas longer, millisecond delays might relinquish the CPU via
 * a process or thread sleep mechanism.  Alternately, the millisecond
 * wait function might just call the microsecond wait function.
 */
#ifndef MICROSECOND_WAIT
#define MICROSECOND_WAIT        MiiMicrosecondWait
extern void     MiiMicrosecondWait(DWORD num_microseconds);
#endif
#ifndef MILLISECOND_WAIT
#define MILLISECOND_WAIT        MiiMillisecondWait
extern void     MiiMillisecondWait(DWORD num_milliseconds);
#endif

/*
 * MiiIoRead() and MiiIoWrite() - export IO_READ() and IO_WRITE()
 * routines for usage by related functions, e.g., "spi.c" routines.
 * The <io_offset> is relative to the base address of the associated
 * MAC or repeater controller device.  The <io_value> read or
 * written is the 32-bit DWORD I/O register value.
 */
extern DWORD    MiiIoRead(PHY  *phy, WORD io_offset);
extern void     MiiIoWrite(PHY *phy, WORD io_offset, DWORD io_value);

#endif /* _MII_H_ */
