/*************************************************************************************************
 *                         PVCS Version Control Information                                      *
 *                                                                                               *
 * $Header:   P:/d878/HomeLAN/Archives/NdisDriver/phy.c_v   1.15   03 Mar 2000 09:49:42   wangmx  $                            
 *                                                                                               *
 ************************************************************************************************/

/*
 *  Module Name: phy.c
 *  Author: Joe Bonaker
 *
 *  Copyright  1997-1999 Conexant Systems, Inc.
 *  All rights reserved.
 *
 *  This software may be used and distributed according to the terms
 *  of the GNU Public License, incorporated herein
 */

/*
 *  Module Description: 100Base-T PHY driver module.  Consists of PHY
 *  primitives supported by underlying "mii.c" MII and adapter/OS
 *  services.  The primary services used are PhyInit() and a periodic
 *  PhyCheck(), the latter of which can be periodically invoked and/or
 *  dispatched upon PHY-to-MAC generated interrupt, depending on
 *  adapter configuration.
 *
 *  Conditional Compilation Switches:
 *
 *     The following conditional compilation switches apply, and may
 *     be overridden by the user at compile-time.
 *
 *     LINK_DOWN_HOP_COUNT - default 5.  When a multiple-device (i.e.,
 *     both HomePNA and MII PHY devices populated) PHY experiences
 *     link down for 5 consecutive calls of PhyCheck(), this routine
 *     may switch or "hop" ports based on the following logic:
 *
 *     select current, "active" port;
 *     if (++link_down_count > LINK_DOWN_HOP_COUNT) {
 *         clear link_down_count;
 *         if (PHY_HWOPTS_PORT_HOPPING) {
 *             select the other, "inactive" port;
 *         } else if (PHY_HWOPTS_MII_PORT_PRIORITY) {
 *             select the MII port;
 *         } else if (PHY_HWOPTS_HPNA_PORT_PRIORITY) {
 *             select the HomePNA port;
 *         } else {
 *             // retain the current, "active" port;
 *         }
 *     }
 *
 *     The PHY_HWOPTS_PORT_HOPPING option gives devices which
 *     cannot function when isolated a chance to achieve a media link.
 *     Some examples are the 1st generation HomePNA PHY devices plus
 *     National, Enable and Level One MII PHY devices.  The
 *     PHY_HWOPTS_HPNA_PORT_PRIORITY option is a good idea when using
 *     MII PHY devices which can function isolated, such as the TDK,
 *     QSI, GEC-Plessey, and ICS devices.  This allows the PHY
 *     Driver to emit MAC-level link test frames on the foreground
 *     "active" HomePNA PHY to determine a soft link indication while
 *     the background "inactive" MII PHY attempts to establish its
 *     hardware link indication.  The PHY_HWOPTS_MII_PORT_PRIORITY
 *     will effectively preclude HomePNA operation unless used in
 *     conjunction with PHY_HWOPTS_PORT_HOPPING as a way of
 *     establishing an initial MII port state, e.g., subsequent to
 *     DriverReset() where the initial state will be the last known
 *     link state.
 *
 *     PORTABILITY_CHECK - default #undefined.  This should only be
 *     #defined when initially porting this module to a new compiler/OS
 *     environment.  This switch enables some simple PhyInit() sizeof()
 *     tests of BYTE, WORD and DWORD types to ensure proper compiler
 *     sizing of register-based fields.  Typically, most compilers
 *     will generate "Unreachable Code" warnings which actually
 *     indicates proper sizing.  Relaxed warnings or older compilers
 *     may warrant a run-time check, for which the PhyInit() function
 *     will return a PHY_STATUS_PORTABILITY_ERROR result.
 *
 * $Name: v2.22 $
 * $Log:   P:/d878/HomeLAN/Archives/NdisDriver/phy.c_v  $
 * 
 *    Rev 1.15   03 Mar 2000 09:49:42   wangmx
 * JoeB,
 * Add Kendin phy support
 * 
 *    Rev 1.14   14 Dec 1999 13:19:16   wangmx
 * MW: 
 * 1)HCT9.5 1c_reset, 1c_fault fix.
 * 2)Cable modem link report, cable modem
 *   media type,
 * 
 *    Rev 1.13   08 Sep 1999 13:22:40   wangmx
 * JC:
 * Fix problem which prevent HCF from reading
 * EEPROM when HLAN is loaded.
 * 
 *    Rev 1.12   25 Aug 1999 16:04:24   wangmx
 * MW:
 * 1)Implement miniport spinlock.
 * 2)Runing phycheck from dispatcher
 * 3)Overhaul Isr and Dpc
 * 
 *    Rev 1.11   10 Aug 1999 15:50:32   chenz
 * change the Ramse X3 floor to 6
 * 
 *    Rev 1.10   04 Aug 1999 15:43:26   wangmx
 * JC:
 * 1) Provide X3 support
 * MW:
 * 1) Add X3 Device hook to NDIS5 inf
 * 2) Provide registry hook for NT4 SoftCRC
 * HPNA Heartbeat, and X3 seclection
 * 
 *    Rev 1.9   22 Jul 1999 17:40:18   wangmx
 * JoeB/MW:
 * Fix MediaConn/DisCon error report:
 * 1)Enable port hopping for all MII device
 * 2)Retain active port before phyint
 * 
 * 3)Change OEMSETNT and NETCN61x inf
 * file to encorporate vender id change
 * 4)Enhance Dispatcher routine
 * 
 *    Rev 1.8   09 Jul 1999 15:26:06   wangmx
 * MWang:
 * 1)Implement rss1161xDispatcher for timer
 * extion
 * 2)6Sec smoothing for XReset link reporting
 * 
 * JoeB:
 * 4 retry before link reporting
 * 
 *    Rev 1.7   29 Jun 1999 16:21:52   chenz
 * Set Ramses FLOOR to 0x10.
 * 
 *    Rev 1.6   22 Jun 1999 15:55:36   wangmx
 * JC:
 * Support Ramses.
 * 
 *    Rev 1.5   07 Jun 1999 10:55:08   wangmx
 *  
 * 
 *    Rev 1.4   May 12 1999 12:06:18   wangmx
 *  
 * Revision 1.23  1999/03/03 07:22:02  bonakeaj
 * - Conexant  NDIS PHY_HWOPTS_ALT_1HZ_HPNA_LINK option
 * Revision 1.22  1999/02/17 16:12:46  bonakeaj
 * - Retains active port to avoid simultaneous MII/HPNA thrashing
 * Revision 1.21  1999/02/16 21:30:58  bonakeaj
 * - Disabled remote PHY cmds, DOS SpiWrite bug, & misc. cleanup
 * Revision 1.20  1998/11/21 17:06:04  bonakeaj
 * - 11623 integrated HomeLAN PHY support
 * Revision 1.19  1998/10/09 22:55:23  bonakeaj
 * - Integrated 1M8 HomePNA with multi-PHY support
 * Revision 1.18  1998/02/12 22:15:46  bonakeaj
 * - RSS 11617 R15 GPIO PHY_HWRESET(), WIN32_DIAG NT 4.0 merge
 * Revision 1.17  1998/02/05 03:08:06  bonakeaj
 * - Preamble supression,  QSI, LXT, GEC, TDK and ICS reg definition
 * Revision 1.16  1998/02/02 22:53:34  bonakeaj
 * - Borland C/C++ NDIS 2.01 (DRV_OS == DOS) support via tools.ide
 * Revision 1.15  1998/01/22 15:40:32  bonakeaj
 * - Davicom PHY and DRV_OS=SCO_MDI support
 * Revision 1.14  1998/01/15 01:28:52  bonakeaj
 * - Added QSI, GEC, Level One & TDK bypass PCS/SCR support
 * Revision 1.13  1998/01/07 20:12:56  bonakeaj
 * - TDK Rev-3 PHY support (need to re-verify against Rev-7 devices)
 * Revision 1.12  1997/11/15 03:50:50  bonakeaj
 * - QSI, Level One & GEC PHY support; no longer default NSC CK25DIS
 * Revision 1.11  1997/11/11 21:54:48  bonakeaj
 * - SAR-140 handle QSI PHY addr changes post-SMI SW Reset
 * Revision 1.10  1997/10/20 21:43:44  bonakeaj
 * - PHY_LINK_DEFAULT and PHY_HWOPTS_FDX_CRS support
 * Revision 1.9  1997/09/16 19:21:26  bonakeaj
 * - SAR-55, PHY HW reset was-500 is-1 mSec, exclusive HW/SW reset only
 * Revision 1.8  1997/09/12 00:16:06  bonakeaj
 * - DRV_OS=LINUX port
 * Revision 1.7  1997/09/03 23:51:14  hamletee
 * Revision 1.6  1997/09/03 20:10:14  bonakeaj
 * - SAR-39 fix whereby PhyEnable() and PhyConnectSet() clear/set
 * - Reg0.Bit12 (BMCR A-N enable) due to NSC DP83840A "feature"
 * Revision 1.5  1997/08/21 21:55:12  kamstrd
 * - Updated copyright headers
 * - Replaced RW116?? with RSS116??
 * - Replaced RW11606 with RSS1161x in some comments
 * Revision 1.4  1997/07/24 21:50:40  bonakeaj
 * - Timing and NSC DP83840 changes for WIN32_DIAG integration
 * Revision 1.3  1997/07/11 22:32:04  bonakeaj
 * - Added PHY_LINK_NEW_MEDIA support, minor type changes for
 * - NDIS 3.1 macApi/phyDrv integration, and fixed PhyInit() bug where
 * - PhyDisable()'s isolate action prevented Auto-Negotiation
 * - completion with the NSC DP83840 PHY device.
 **********************************************************************/

#include "phy.h"
#include "mii.h"
#include "spi.h"
#include "phyregs.h"

#if (DRV_OS == LINUX)  
#ifdef __KERNEL__


#include <linux/config.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/version.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#ifdef VER_CTL
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS			/* force it on */
#endif
#endif

#endif
#endif

#ifndef LINK_DOWN_HOP_COUNT
#define LINK_DOWN_HOP_COUNT     5       /* 10 seconds @ 0.5 Hz */
#endif

/* Local (static) function prototypes */
static void     ParsePhyHwConfig(PHY *phy);
static void     SetPhyHwOptions(PHY *phy, DWORD hw_opts);

static PHY_STATUS RamseInit (PHY *phy);
static void DisableRamse(PHY *phy);
static void EnableRamse(PHY *phy);

/* 
 * Local an_lookup[] table maps BMSR and ANAR hardware to internal,
 * soft PHY_LINK_* (or associated PHY_CONN_* alias) bit assignments.
 * The table is ordered in decreasing Auto-Negotiation priority per
 * IEEE 802.3u and 802.3y, and is terminated by a null (zero) record.
 * Searches may be keyed on any of the fields, however, <phy_link>
 * keys should mask out non-PHY_LINK_TECHNOLOGY bits, otherwise the
 * search may fail to achieve a match.
 *
 * This table may have to be extended to include an SNMP MIB field,
 * which contains yet another enumeration of the PHY technology types.
 */
typedef struct {
    WORD            bmsr;       /* BMSR HW capability bit */
    WORD            anar;       /* ANAR/ANLPAR HW bit */
    DWORD           phy_link;   /* internal PHY_LINK_* state */
}                AN_LOOKUP;
static AN_LOOKUP an_lookup[] = {
    {BMSR_100BASE_T2FD, ANAR_100BASE_T2FD,
     PHY_LINK_100BASE_T2FD | PHY_LINK_100 | PHY_LINK_FDX},

    {BMSR_100BASE_TXFD, ANAR_100BASE_TXFD,
     PHY_LINK_100BASE_TXFD | PHY_LINK_100 | PHY_LINK_FDX},

    {BMSR_100BASE_T2,   ANAR_100BASE_T2,
     PHY_LINK_100BASE_T2   | PHY_LINK_100},

    {BMSR_100BASE_T4,   ANAR_100BASE_T4,
     PHY_LINK_100BASE_T4   | PHY_LINK_100},

    {BMSR_100BASE_TX,   ANAR_100BASE_TX,
     PHY_LINK_100BASE_TX   | PHY_LINK_100},

    {BMSR_10BASE_TFD,   ANAR_10BASE_TFD,
     PHY_LINK_10BASE_TFD   | PHY_LINK_10  | PHY_LINK_FDX},

    {BMSR_10BASE_T,     ANAR_10BASE_T,
     PHY_LINK_10BASE_T     | PHY_LINK_10},

    /* null-terminating record */
    {0, 0, 0}
};

/*
 * HomePNA link test frame template for packet-based soft link
 * determination.  The frame type is an LLC Test command.
 */
static char     hpna_link_test_frame_template[] = {
    /* MAC DestAddr */  "\x00\x00\x00\x00\x00\x00"
    /* MAC SrcAddr */   "\x00\x00\x00\x00\x00\x00"
    /* MAC Length */    "\x00\x00"
    /* LLC PDU */       "\x00\x00\xF3"
    /* LLC info */      "HomePNA (version 1.1) link heartbeat frame"
};

/***********************************************************************
 * PhyInit() - Initialize PHY and attach to media (w/optional override)
 **********************************************************************/
PHY_STATUS
PhyInit(DWORD  adapter,         /* MAC device I/O base address */
        PHY   *phy,             /* allocated PHY data structure space */
        BYTE   phy_addr,        /* MII PHY address or wildcard */
        DWORD  connect,         /* PhyConnectSet() directive */
        DWORD  hw_options,      /* zero or more PHY_HWOPTS_* values */
        DWORD *soft_nar)        /* used with PHY_HWOPTS_MAC_PORT_MGMT */
{
    BYTE        first_addr;
    BYTE        last_addr;
    PHY_STATUS  mii_status;
    PHY_STATUS  hpna_status;
    PHY_STATUS  status;
    int         num_restarts;
    int         n;
    WORD        bmsr;
    WORD        ctrl;

#ifdef PORTABILITY_CHECK        /* ################################## */
    /* Not a bad idea during initial PHY driver porting effort */
    if ((sizeof(BYTE) != 1) ||
            (sizeof(WORD) != 2) || (sizeof(DWORD) != 4)) {
        return (PHY_STATUS_PORTABILITY_ERROR);
    }
#endif  /* PORTABILITY_CHECK */ /* ################################## */

    /*
     * Set-up <phy> data and perform optional pre-MII search HW reset,
     * which may be required only in the case of unusual adapter design.
     * This case has been included to satisfy the RSS11606 test adapter
     * board, wherein a MAC general purpose I/O (GPIO) pin holds the PHY
     * in hardware reset.  Unless programmed to release the PHY prior to
     * searching the MII bus, the subsequent MII search will fail, as
     * the PHY obviously cannot respond when held in hardware reset.
     *
     * Bug warning:  unless there's a crafty HW design, this could
     * cause unexpected failures in a multiple-PHY environment if
     * PHY_ADDR_ANY or PHY_ADDR_NEXT() is specified to search for a
     * responding PHY.  For now, we'll simply do as we're told.  This
     * should probably be moved into the search loop below, and rely
     * upon a PHY_HWRESET() routine that parses the phy_addr argument...
     */
    for (n = 0; n < sizeof(*phy); n++) {
        /* portability: zero PHY struct w/o library memset() call */
        ((BYTE *) phy)[n] = 0;
    }
    {
        /* HomePNA PHY_CALLBACK link test frame copy */
        for (n = 0; n < sizeof(hpna_link_test_frame_template); n++) {
            if (n >= sizeof(phy->callback.test_frame)) {
                break;
            }
            phy->callback.test_frame[n] =
                (BYTE) hpna_link_test_frame_template[n]; 
        }
        phy->callback.test_frame_size = (DWORD) n;
        phy->callback.test_frame[12] =          /* Ether Length MSB */
            (BYTE) ((phy->callback.test_frame_size - 14) >> 8);
        phy->callback.test_frame[13] =          /* Ether Length LSB */
            (BYTE) (phy->callback.test_frame_size - 14);
    }
    if (hw_options & PHY_HWOPTS_READ_ONLY) {
        phy->read_only = TRUE;
    }
    if (hw_options & PHY_HWOPTS_MAC_PORT_MGMT) {
        /* Save pointer to caller's soft shadow copy of MAC Reg 6 */
        phy->soft_nar = soft_nar;
    }
    phy->adapter = adapter;
    if (hw_options & PHY_HWOPTS_HWRESET_MASK) {
        phy->addr = phy_addr;           /* if needed by HW reset */
        PHY_HWRESET(phy, hw_options);
    }

    /*
     * Search for HomePNA devices.  Assume a device is found if
     * SPI register zero <ctrl> is not all zeroes or all ones.
     * Precedence: disable the internal device (PHY_PORT_HPNA_INT)
     * if an external device is also found.
     */
    hpna_status = PHY_STATUS_NOT_FOUND;
    if (!(hw_options & PHY_HWOPTS_NO_HPNA_SEARCH)) {
        /* Internal (RSS 11623) HomePNA device search */
        SpiInit(phy, PHY_PORT_HPNA_INT, 0, 0);
        ctrl = SpiReadWord(phy, HPNA_CONTROL);
        if ((ctrl == 0x0000) || (ctrl == 0xFFFF)) {
            /* Assume nothing there */
            phy->hpna_port = PHY_PORT_NONE;
        } else {
            hpna_status = PHY_STATUS_OK;
            phy->devices |= PHY_PORT_HPNA_INT;
            SpiWriteEnable(phy);
        }

        /* External Ramses/11623 HomePNA device search */
        if (!(hw_options & PHY_HWOPTS_NO_HPNA_EXT_SEARCH)) {
            phy->hw_options = hw_options;       // Init
            SpiInit(phy,
                    PHY_PORT_HPNA_EXT,
                    hw_options,
                    0);
            ctrl = SpiReadWord(phy, HPNA_CONTROL);
            if ((ctrl == 0x0000) || (ctrl == 0xFFFF)) {
                phy->hpna_port = (phy->devices & PHY_PORT_HPNA_INT)
                    ? PHY_PORT_HPNA_INT : PHY_PORT_NONE;
                phy->hw_options = 0;
            } else {
                hpna_status = PHY_STATUS_OK;
                phy->devices |= PHY_PORT_HPNA_EXT;
                SpiWriteEnable(phy);
            }
        }

#ifdef OBSOLETE_11617   /* D500 board 11617 no longer supported */
        /* Skip search 11617 if 11623 or Ramses are found. */
        if (hpna_status != PHY_STATUS_OK) {
            /* External (RSS 11617-X4) HomePNA device search */
            if (!(hw_options & PHY_HWOPTS_NO_HPNA_EXT_SEARCH)) {
                SpiInit(phy,
                        PHY_PORT_HPNA_EXT,
                        (hw_options & PHY_HWOPTS_HWRESET_MASK)
                            ? HPNA_GPIO_RESET : 0,
                        SPI_GPIO_DEFAULT);
                ctrl = SpiReadWord(phy, HPNA_CONTROL);
                if ((ctrl == 0x0000) || (ctrl == 0xFFFF)) {
                    phy->hpna_port = (phy->devices & PHY_PORT_HPNA_INT)
                        ? PHY_PORT_HPNA_INT : PHY_PORT_NONE;
                } else {
                    hpna_status = PHY_STATUS_OK;
                    phy->devices |= PHY_PORT_HPNA_EXT;
                    SpiWriteEnable(phy);
                }
            }
        }
#endif  /* OBSOLETE_11617 */

    }

    if ((phy->devices & (PHY_PORT_HPNA_INT | PHY_PORT_HPNA_EXT)) ==
            (PHY_PORT_HPNA_INT | PHY_PORT_HPNA_EXT)) {
        /* Two HomePNA devices: external supercedes internal */
        phy->hpna_port = PHY_PORT_HPNA_INT;
        PhyDisable(phy, PHY_PORT_HPNA_INT);
        phy->hpna_port = PHY_PORT_HPNA_EXT;
        phy->devices &= ~PHY_PORT_HPNA_INT;     /* deleted */
    }

    /* Fix HomePNA noise thresholds, and disable remote commands */
    if (phy->hpna_port != PHY_PORT_NONE) {
        phy->ramse_ctrl_reg = 0;
        phy->hw_options = hw_options;       // Init
        if (RamseInit(phy) != PHY_STATUS_OK) {
            /* Not Ramses */
            SpiWriteByte(phy, HPNA_NSE_FLOOR,   HPNA_NSE_FLOOR_SWOVERRIDE);
            SpiWriteByte(phy, HPNA_NSE_CEILING, HPNA_NSE_CEILING_SWOVERRIDE);
        }
        ctrl = SpiReadWord(phy, HPNA_CONTROL);
        ctrl |= HPNA_CONTROL_IGNORE_REMOTE_CMD;
        if (hw_options & PHY_HWOPTS_HPNA_CMD_ENABLE) {
            ctrl &= ~HPNA_CONTROL_IGNORE_REMOTE_CMD;
        }
        SpiWriteWord(phy, HPNA_CONTROL, ctrl);
    }

    /*
     * MII search, reset and optionally disable (isolate) PHY.  The
     * search logic simply reads the BMSR, for which a responding PHY
     * should return a value that is non-zero and not all one's.
     * Otherwise, assume there isn't an MII-capable PHY at <phy_addr>.
     * The ANAR probe is repeated after post-HW and/or software reset to
     * guard against the PHY losing its identity (address) upon reset.
     *
     * If the caller specified PHY_ADDR_ANY or PHY_ADDR_NEXT(phy#), 
     * we'll walk the list of valid PHY addresses until the first/next
     * responding PHY is detected.  Some PHY devices have side-effects
     * associated with PHY_ADDR_MIN (zero) -- this address is often
     * (erroneously) treated as a broadcast by some PHYs, others
     * won't display vendor-specific registers when reading from
     * address-0, etc.  As such, we'll initiate a PHY_ADDR_ANY search
     * from (PHY_ADDR_MIN + 1) on up to PHY_ADDR_MAX, concluding the
     * search using PHY_ADDR_MIN only if no other addresses are found.
     * This can be overriden by the PHY_HWOPTS_MII_SEARCH_0 <hw_options>
     * switch, which initiates a search from address 0..31, rather
     * than the default 1..31 followed by 0.
     */
    mii_status = PHY_STATUS_NOT_FOUND;
    if (phy_addr == PHY_ADDR_ANY) {
        first_addr = PHY_ADDR_MIN;      /* full MII search, below */
        last_addr  = PHY_ADDR_MAX;
        phy_addr   = first_addr + 1;    /* start search at address-1 */
        if (hw_options & PHY_HWOPTS_MII_SEARCH_0) {
            --phy_addr;                 /* start search at address-0 */
        }
    } else if (/* (PHY_ADDR_MIN <= phy_addr) && */
               (phy_addr     <= PHY_ADDR_MAX)) {
        first_addr = phy_addr;          /* single MII search, below */
        last_addr  = phy_addr;
    } else {
        first_addr = -phy_addr;         /* partial? MII search per */
        last_addr  = PHY_ADDR_MAX;      /* phy.h PHY_ADDR_NEXT() */
        phy_addr   = first_addr;
    }
    num_restarts = 1 + (last_addr - first_addr);
    while (phy_addr <= last_addr) {
        /* Bail out if we're excluding MII searches */
        if (hw_options & PHY_HWOPTS_NO_MII_SEARCH) {
            break;
        }

        phy->addr = phy_addr++;
        // if (IS_ANAR_PROTOCOL_IEEE_802_3(MiiRead(phy, MII_ANAR)))
        if (((bmsr = MiiRead(phy, MII_BMSR)) != 0x0000) &&
                (bmsr != 0xFFFF)) {
            /* Assume PHY found, so nominally perform post-SW reset */
            mii_status = PHY_STATUS_OK;
            if (!(hw_options & PHY_HWOPTS_DONT_SWRESET)) {
                mii_status = PhyReset(phy); /* nominal software reset */
            }

            /* Terminate upon successful reset(s) or excess restarts */
            if ((mii_status == PHY_STATUS_OK) &&
                    // IS_ANAR_PROTOCOL_IEEE_802_3(MiiRead(phy, MII_ANAR))
                    ((bmsr = MiiRead(phy, MII_BMSR)) != 0x0000) &&
                    (bmsr != 0xFFFF)) {
                phy->devices |= PHY_PORT_MII_DEVICE;
                phy->mii_port = PHY_PORT_MII;
                phy->id = (((DWORD) MiiRead(phy, MII_PHYIDR1) << 16) |
                           ((DWORD) MiiRead(phy, MII_PHYIDR2)));
                break;

            } else if (--num_restarts <= 0) {
                if (mii_status == PHY_STATUS_OK) {
                    mii_status = PHY_STATUS_NOT_FOUND;
                }
                break;
            }

            /* Restart search on reset error or apparent addr change */
            if (mii_status == PHY_STATUS_OK) {
                mii_status = PHY_STATUS_NOT_FOUND;
            }
            phy_addr = first_addr;
            if ((first_addr == PHY_ADDR_MIN) &&
                    (last_addr == PHY_ADDR_MAX) &&
                    !(hw_options & PHY_HWOPTS_MII_SEARCH_0)) {
                // phy_addr++;          /* full search, avoid addr-0? */
            }
        }

        /* Go back to addr-0 once full MII search 1..31 fails */
        if ((phy_addr == PHY_ADDR_MAX) &&
                (first_addr == PHY_ADDR_MIN) &&
                !(hw_options & PHY_HWOPTS_MII_SEARCH_0)) {
            phy_addr = first_addr = last_addr = PHY_ADDR_MIN;
        }
    }

    if ((hpna_status != PHY_STATUS_OK) &&
            (mii_status != PHY_STATUS_OK)) {
        return (mii_status);
    }
    status = PHY_STATUS_OK;

    /*
     * Parse the SPI and/or MII HW config info, and establish
     * any caller-specified HW options.  Establish the PHY connect
     * directive, and then nominally attach to the media (just in
     * case the default PHY state is disabled/isolated).
     *
     * Side-effect:  MDIO <preamble_suppression>, if supported by
     * the device and not overridden by the PHY_HWOPTS_MII_PREAMBLE
     * option, will be in effect subsequent to the ParsePhyHwConfig()
     * call.
     */
    ParsePhyHwConfig(phy);
    SetPhyHwOptions(phy, hw_options);
    /* Possibly the following block to PhyConnectSet()? */
    {   
        if (phy->devices & PHY_PORT_MII) {
            MII_CLR_MASK(phy, MII_BMCR, (BMCR_RESTART_AUTONEG |
                                         BMCR_POWERDOWN       |
                                         BMCR_LOOPBACK        |
                                         BMCR_COLLISION_TEST));
        }
        if (hw_options & PHY_HWOPTS_ISOLATE) {
            (void) PhyDisable(phy, phy->devices);
        }
    }
    if ((status = PhyConnectSet(phy, connect)) == PHY_STATUS_OK) {
        phy->initialized = PHY_INIT_CODE;
        phy->hw_options &= ~PHY_HWOPTS_ISOLATE;
    }
//JC debug
#if 0
#if (DRV_OS == LINUX)  
#ifdef __KERNEL__

    printk ("phy[0] = 0x%x\n", MiiRead(phy, 0));
    printk ("phy[1] = 0x%x\n", MiiRead(phy, 1));
    printk ("phy[2] = 0x%x\n", MiiRead(phy, 2));
    printk ("phy[3] = 0x%x\n", MiiRead(phy, 3));
    printk ("phy[4] = 0x%x\n", MiiRead(phy, 4));
    printk ("phy[17] = 0x%x\n", MiiRead(phy, 17));
#endif
#endif
#endif

    return (status);
}

/***********************************************************************
 * ParsePhyHwConfig() - Process PHY HW configuration options
 **********************************************************************/
static void
ParsePhyHwConfig(PHY *phy)
{
    WORD        ctrl;
    WORD        bmcr;
    WORD        bmsr;
    WORD        anar;
    DWORD       mode;
    AN_LOOKUP  *lookup;

    phy->media.capability = 0;
    phy->media.connect = 0;

    /*-------------------------*
     * (1) Examine HomePNA PHY
     *-------------------------*/

    if (phy->hpna_port != PHY_PORT_NONE) {
        phy->media.capability |= PHY_LINK_HPNA_TECHNOLOGY;
        ctrl = SpiReadWord(phy, HPNA_CONTROL);
        switch (ctrl & (HPNA_CONTROL_HIGH_POWER |
                        HPNA_CONTROL_HIGH_SPEED)) {
            default:
            case HPNA_CONTROL_HIGH_SPEED:       /* w/low power */
                phy->media.connect |= PHY_CONN_1HPNA;
                break;

            case (HPNA_CONTROL_HIGH_SPEED | HPNA_CONTROL_HIGH_POWER):
                phy->media.connect |= PHY_CONN_1HPNA_HIPWR;
                break;

            case 0:                             /* low speed, low power */
                phy->media.connect |= PHY_CONN_1HPNA_LOSPD;
                break;

            case HPNA_CONTROL_HIGH_POWER:       /* low speed */
                phy->media.connect |= PHY_CONN_1HPNA_HIPWR_LOSPD;
                break;
        }

        /* Check for isolated (tri-stated) HomePNA PHY */
        if (ctrl & HPNA_CONTROL_POWERDOWN) {
            phy->hpna_isolated = TRUE;
        }

        if (phy->mii_port == PHY_PORT_NONE) {
            /* Coerce soft non-MII Auto-Negotiation capability */
            phy->media.capability |= PHY_LINK_AUTONEG;
            phy->media.connect    |= PHY_CONN_AUTONEG;
            return;
        }
    }

    /*-------------------------*
     * (2) Examine MII PHY
     *-------------------------*/

    /* Determine media capabilities based on BMSR */
    bmsr = MiiRead(phy, MII_BMSR);
    if (bmsr & BMSR_AUTONEG_ABILITY) {
        phy->media.capability |= PHY_LINK_AUTONEG;
    }
    for (lookup = &an_lookup[0]; lookup->bmsr != 0; lookup++) {
        if (bmsr & lookup->bmsr) {
            phy->media.capability |=
                (lookup->phy_link & PHY_LINK_TECHNOLOGY);
        }
    }

    /*
     * Determine adapter (NIC) unique HW configuration strap
     * settings based on BMCR (and ANAR if Auto-Negotiation is enabled).
     */
    bmcr = MiiRead(phy, MII_BMCR);
    if (bmcr & BMCR_AUTONEG_ENABLE) {
        /* Read default Auto-Negotiation configuration per ANAR */
        phy->media.connect |= PHY_CONN_AUTONEG;
        anar = MiiRead(phy, MII_ANAR);
        for (lookup = &an_lookup[0]; lookup->anar != 0; lookup++) {
            if (anar & lookup->anar) {
                phy->media.connect |=
                    (lookup->phy_link & PHY_LINK_TECHNOLOGY);
            }
        }
    } else {
        /* Infer default forced mode configuration per BMCR */
        mode = 0;
        if (bmcr & BMCR_SPEED_100) {
            mode |= PHY_LINK_100;
        } else {
            mode |= PHY_LINK_10;
        }
        if (bmcr & BMCR_DUPLEX_MODE) {
            mode |= PHY_LINK_FDX;
        }
        for (lookup = &an_lookup[0]; lookup->phy_link != 0; lookup++) {
            if ((phy->media.capability & lookup->phy_link) &&
                    (mode == (lookup->phy_link &
                              (PHY_LINK_SPEED | PHY_LINK_FDX)))) {
                phy->media.connect |= lookup->phy_link;
                break;
            }
        }
    }

    /* Check for isolated (tri-stated) MII PHY */
    if (bmcr & BMCR_ISOLATE) {
        phy->mii_isolated = TRUE;
    }

    /* 
     * Finally, set <preamble_suppression> if suported by the device.
     * This can subsequently be overriden by SetPhyHwOptions()
     * via the PHY_HWOPTS_MII_PREAMBLE directive, which may be used
     * in a multi-PHY environment or in case a device's BMSR register
     * falsely indicates preamble suppression capability.
     */
    if (bmsr & BMSR_PREAMBLE_SUPPRESS) {
        phy->preamble_suppression = TRUE;
    }
}

/***********************************************************************
 * SetPhyHwOptions() - Configure non-default MII PHY hardware features
 **********************************************************************/
static void
SetPhyHwOptions(PHY *phy, DWORD hw_opts)
{
    phy->hw_options = hw_opts;

    /* Force MDIO preamble, i.e., explicitly disable suppression */
    if (hw_opts & PHY_HWOPTS_MII_PREAMBLE) {
        phy->preamble_suppression = FALSE;
    }

    if (PHYID_NSC_DP83840(phy)) {
        /* Force node or repeater mode, which affects CRS signal */
        if (hw_opts & PHY_HWOPTS_REPEATER) {
            MII_SET_BIT(phy, MII_PCR, PCR_REPEATER_MODE);
        } else {
            MII_CLR_BIT(phy, MII_PCR, PCR_REPEATER_MODE);
        }

        /* Disable CLK25 output line, reducing noise when unused */
        if (hw_opts & PHY_HWOPTS_CK25DIS) {
            MII_SET_BIT(phy, MII_PCR, PCR_CLK25DIS);
        }

        /* Bypass portions of the transmitter and receiver coding */
        if (hw_opts & PHY_HWOPTS_TX_BPALIGN) {
            MII_SET_BIT(phy, MII_LBREMR, LBREMR_BP_ALIGN);
        }
        if (hw_opts & PHY_HWOPTS_TX_BPSCR) {
            MII_SET_BIT(phy, MII_LBREMR, LBREMR_BP_SCR);
        }
        if (hw_opts & PHY_HWOPTS_TX_BP4B5B) {
            MII_SET_BIT(phy, MII_LBREMR, LBREMR_BP_4B5B);
        }

        /* Assert FDX CRS during transmit, not receive (default) */
        if (hw_opts & PHY_HWOPTS_FDX_CRS) {
            if (PHYID_REV(phy) >= 1) {
                /* New NSC DP83840 Rev A (revision code 1) feature */
                MII_SET_BIT(phy, MII_LBREMR, LBREMR_ALT_CRS);
            }
        }

    } else if (PHYID_QSI_6612(phy)) {
        /* Bypass portions of the transmitter and receiver coding */
        if (hw_opts & PHY_HWOPTS_TX_BPALIGN) {
            /* No QSI support for bypassing the alignment function */
        }
        if (hw_opts & PHY_HWOPTS_TX_BPSCR) {
            MII_SET_BIT(phy, MII_R31, QSI_R31_SCR_DISABLE);
        }
        if (hw_opts & PHY_HWOPTS_TX_BP4B5B) {
            MII_CLR_BIT(phy, MII_R31, QSI_R31_4B5BEN);
        }

    } else if (PHYID_LXT_970(phy)) {
        /* Bypass portions of the transmitter and receiver coding */
        if (hw_opts & PHY_HWOPTS_TX_BPALIGN) {
            /* No Level One support for bypassing alignment function */
        }
        if (hw_opts & PHY_HWOPTS_TX_BPSCR) {
            MII_SET_BIT(phy, MII_R19, LXT_R19_SCR_BYPASS);
        }
        if (hw_opts & PHY_HWOPTS_TX_BP4B5B) {
            MII_SET_BIT(phy, MII_R19, LXT_R19_4B5B_BYPASS);
        }


    } else if (PHYID_GEC_NWK936(phy)) {
        /* Bypass portions of the transmitter and receiver coding */
        if (hw_opts & PHY_HWOPTS_TX_BPALIGN) {
            MII_SET_BIT(phy, MII_R24, GEC_R24_BPALIGN);
        }
        if (hw_opts & PHY_HWOPTS_TX_BPSCR) {
            MII_SET_BIT(phy, MII_R24, GEC_R24_BPSCR);
        }
        if (hw_opts & PHY_HWOPTS_TX_BP4B5B) {
            MII_SET_BIT(phy, MII_R24, GEC_R24_BPENC);
        }

        /* Assert FDX CRS during transmit, not receive (default) */
        if (!(hw_opts & PHY_HWOPTS_FDX_CRS)) {
            MII_SET_BIT(phy, 24, 0x0800);
        }

    } else if (PHYID_TDK_2120(phy)) {
        /* Force node or repeater mode, which affects CRS signal */
        if (hw_opts & PHY_HWOPTS_REPEATER) {
            MII_SET_BIT(phy, MII_R16, TDK_R16_RPTR);
        } else {
            MII_CLR_BIT(phy, MII_R16, TDK_R16_RPTR);
        }

        /*
         * Disable CLK25 output line, reducing noise when unused.  Due
         * to Auto-Negotiation side-effects, the TDK RXCC option is
         * now performed by PhyCheck() on link up and down transitions.
         *
         * if (hw_opts & PHY_HWOPTS_CK25DIS) {
         *     MII_SET_BIT(phy, MII_R16, TDK_R16_RXCC);
         * }
         */

        /* Bypass portions of the transmitter and receiver coding */
        if (hw_opts & (PHY_HWOPTS_TX_BPSCR | PHY_HWOPTS_TX_BP4B5B)) {
            /* TDK PHY doesn't have separate PCS, scrambler bypass */
            MII_SET_BIT(phy, MII_R16, TDK_R16_PCSBP);
        }
    }
}

/***********************************************************************
 * PhyReset() - Issue PHY software reset, waiting 500 mSec to complete
 **********************************************************************/
PHY_STATUS
PhyReset(PHY *phy)
{
    PHY_STATUS  phy_status = PHY_STATUS_RESET_ERROR;
    int         n;

    /*
     * Set BMCR Reset bit, expected to auto-clear in 500 milliseconds
     * Some devices may require a 500-microsecond guard against serial
     * MII access, so we'll delay after setting the reset bit and after
     * detecting auto-clear, just to be safe.
     */
    MII_SET_BIT(phy, MII_BMCR, BMCR_RESET);
    MICROSECOND_WAIT(500);
    for (n = 0; n < 50; n++) {
        if (!(MiiRead(phy, MII_BMCR) & BMCR_RESET)) {
            phy_status = PHY_STATUS_OK;
            break;
        }
        MILLISECOND_WAIT(10);
    }
    MICROSECOND_WAIT(500);
    return (phy_status);
}


/***********************************************************************
 * PhyDisable() - Disable PHY by asserting BMCR isolate command
 **********************************************************************/
PHY_STATUS
PhyDisable(PHY *phy, PHY_PORT port)
{
    WORD        ctrl;

    /* (1) Validate the selected port(s) */
    if (port & PHY_PORT_ACTIVE)    port |= phy->active_port;
    if (port & PHY_PORT_INACTIVE)  port |= phy->inactive_port;
    if (port & PHY_PORT_ENABLED)   port |= phy->enabled_port;
    if (port & PHY_PORT_DISABLED)  port |= phy->disabled_port;
    port &= phy->devices;

    /* (2) MII PMD isolation (TP tri-state) */
    if (port & PHY_PORT_MII_PMD) {
        if (PHYID_QSI_6612(phy)) {
            MII_SET_BIT(phy, MII_R31, QSI_R31_TX_ISOLATE);
        } else if (PHYID_LXT_970(phy)) {
            MII_SET_BIT(phy, MII_R19, LXT_R19_TX_DISCONNECT);
        }
    }

    /* (3) MII data isolation (tri-state) */
    if ((port & PHY_PORT_MII) && !phy->mii_isolated) {
        phy->mii_isolated = TRUE;
        MII_SET_BIT(phy, MII_BMCR, BMCR_ISOLATE);
    }

    /* (4) Internal (RSS 11623) or external HomePNA device */
    if ((port & (PHY_PORT_HPNA_INT |
                 PHY_PORT_HPNA_EXT)) && !phy->hpna_isolated) {

        phy->hpna_isolated = TRUE;
        /*
         *  Disable Ramses if Ramses is detected
         */
        DisableRamse(phy);

        ctrl = SpiReadWord(phy, HPNA_CONTROL);
        ctrl |= HPNA_CONTROL_POWERDOWN;
        SpiWriteWord(phy, HPNA_CONTROL, ctrl);
    }

    return (PHY_STATUS_OK);
}

/***********************************************************************
 * PhyEnable() - (Re)Enable PHY by deasserting BMCR isolate command
 **********************************************************************/
PHY_STATUS
PhyEnable(PHY *phy, PHY_PORT port)
{
    WORD        bmcr;
    WORD        ctrl;

    /* (1) Validate the selected port(s) */
    if (port & PHY_PORT_ACTIVE)    port |= phy->active_port;
    if (port & PHY_PORT_INACTIVE)  port |= phy->inactive_port;
    if (port & PHY_PORT_ENABLED)   port |= phy->enabled_port;
    if (port & PHY_PORT_DISABLED)  port |= phy->disabled_port;
    port &= phy->devices;

    /* (2) MII PMD de-isolation (TP non tri-state) */
    if (port & PHY_PORT_MII_PMD) {
        if (PHYID_QSI_6612(phy)) {
            MII_CLR_BIT(phy, MII_R31, QSI_R31_TX_ISOLATE);
        } else if (PHYID_LXT_970(phy)) {
            MII_CLR_BIT(phy, MII_R19, LXT_R19_TX_DISCONNECT);
        }
    }

    /* (3) MII data de-isolation (no longer tri-stated) */
    if ((port & PHY_PORT_MII) && phy->mii_isolated) {
        phy->mii_isolated = FALSE;
        bmcr = MiiRead(phy, MII_BMCR) & ~BMCR_ISOLATE;
        if (phy->media.connect & PHY_CONN_AUTONEG) {
            bmcr |= (BMCR_AUTONEG_ENABLE | BMCR_RESTART_AUTONEG);
            /*
             * Side-effect:  some PHYs (well, at least one vendor's)
             * require the BMCR_AUTONEG_ENABLE bit (reg 0.12) be cleared
             * then set in the case where the device is coming up out
             * of reset or isolated state, otherwise, Auto-Negotiation
             * won't be initiated.  We'll temporarily clear A-N bits,
             * which will then be immediately be set, below.
             */
            if (PHYID_NSC_DP83840(phy) ||
                    (phy->hw_options & PHY_HWOPTS_ACTIVE_AN_RESTART)) {
                MiiWrite(phy, MII_BMCR,
                         (WORD) (bmcr & ~(BMCR_AUTONEG_ENABLE |
                                          BMCR_RESTART_AUTONEG)));
            }
        }
        MiiWrite(phy, MII_BMCR, bmcr);
    }

    /* (4) Internal (RSS 11623) or external HomePNA device */
    if ((port & (PHY_PORT_HPNA_INT |
                 PHY_PORT_HPNA_EXT)) && phy->hpna_isolated) {
        phy->hpna_isolated = FALSE;
        ctrl = SpiReadWord(phy, HPNA_CONTROL);
        ctrl &= ~HPNA_CONTROL_POWERDOWN;
        SpiWriteWord(phy, HPNA_CONTROL, ctrl);
        /*
         *  Enable Ramses if Ramses is detected
         */
        EnableRamse(phy);
    }

    return (PHY_STATUS_OK);
}

/***********************************************************************
 * PhyConnectGet() - Get PHY connect setting(s)
 **********************************************************************/
DWORD
PhyConnectGet(PHY *phy)
{
    return (phy->media.connect);
}

/***********************************************************************
 * PhyConnectSet() - Execute PHY connect settings
 **********************************************************************/
PHY_STATUS
PhyConnectSet(PHY *phy, DWORD connect)
{
    WORD        bmcr;
    WORD        anar;
    AN_LOOKUP  *lookup;
    WORD        ctrl;

    /* (1) Establish/adjust the <connect> directive */
    switch (connect & (PHY_CONN_AUTONEG | PHY_CONN_TECHNOLOGY)) {
        default:
            /* Use non-zero <connect> directive as-is */
            break;

        case PHY_CONN_CURRENT:
            /* Zero: use HW default or previous setting */
            connect = phy->media.connect; break;

        case PHY_CONN_AUTONEG:
            /* A-N bit only: Auto-Negotiate HW default or previous */
            connect = phy->media.connect | PHY_CONN_AUTONEG; break;
    }

    if (phy->hw_options & PHY_HWOPTS_STRICT_PARM_CHECKING) {
        /* Strict checking:  A-N and full media agreement required */
        if ((connect & (PHY_CONN_AUTONEG | PHY_CONN_TECHNOLOGY)) !=
                (connect & phy->media.capability)) {
            return (PHY_STATUS_UNSUPPORTED_MEDIA);
        }

    } else {
        /* Relaxed checking:  A-N and partial media agreement needed */
        if (((connect & PHY_CONN_AUTONEG)
             && !(phy->media.capability & PHY_CONN_AUTONEG)) ||
            ((connect & PHY_CONN_TECHNOLOGY) &&
             !(connect & phy->media.capability &
               PHY_CONN_TECHNOLOGY))) {
            return (PHY_STATUS_UNSUPPORTED_MEDIA);
        }
        /* Truth in advertising:  shave unsupported media settings */
        connect &= (phy->media.capability | ~PHY_CONN_TECHNOLOGY);
    }

    /* (2) HomePNA device: force highest-priority setting */
    if (connect & PHY_CONN_HPNA_TECHNOLOGY) {
        ctrl = SpiReadWord(phy, HPNA_CONTROL);
        ctrl &= ~(HPNA_CONTROL_HIGH_POWER | HPNA_CONTROL_HIGH_SPEED);

        if (connect & PHY_CONN_1HPNA) {
            /* Priority #1 - high speed, low  power */
            connect &= ~(PHY_CONN_HPNA_TECHNOLOGY &
                         ~PHY_CONN_1HPNA);
            ctrl |= HPNA_CONTROL_HIGH_SPEED;

        } else if (connect & PHY_CONN_1HPNA_HIPWR) {
            /* Priority #2 - high speed, high power */
            connect &= ~(PHY_CONN_HPNA_TECHNOLOGY &
                         ~PHY_CONN_1HPNA_HIPWR);
            ctrl |= (HPNA_CONTROL_HIGH_POWER |
                     HPNA_CONTROL_HIGH_SPEED);

        } else if (connect & PHY_CONN_1HPNA_LOSPD) {
            /* Priority #3 - low  speed, low  power */
            connect &= ~(PHY_CONN_HPNA_TECHNOLOGY &
                         ~PHY_CONN_1HPNA_LOSPD);

        } else if (connect & PHY_CONN_1HPNA_HIPWR_LOSPD) {
            /* Priority #4 - low  speed, high power */
            connect &= ~(PHY_CONN_HPNA_TECHNOLOGY &
                         ~PHY_CONN_1HPNA_HIPWR_LOSPD);
            ctrl |= HPNA_CONTROL_HIGH_POWER;
        }
        SpiWriteWord(phy, HPNA_CONTROL, ctrl);
        SpiWriteWord(phy, HPNA_IMASK, 0);       /* no wake-on-link */
        /* Ramses has a bug, which is Hi/Lo power mode can't be set/reset
           by HPNA_CONTROL */
        if (phy->hpna_port == PHY_PORT_HPNA_EXT && phy->gpio_spi_phy_csn == 0) {
            WORD wAfeCtl;

            wAfeCtl = SpiReadWord(phy, HPNA_AFE_CONTROL);
            if (ctrl & HPNA_CONTROL_HIGH_POWER)
                wAfeCtl |= HPNA_CONTROL_HIGH_POWER;
            else
                wAfeCtl &= ~HPNA_CONTROL_HIGH_POWER;
            SpiWriteWord(phy, HPNA_AFE_CONTROL, wAfeCtl);
        }
    }

    /* (3) MII device: Auto-Negotiate or force highest capability */
    if (connect & PHY_CONN_MII_TECHNOLOGY) {

        /* Read BMCR, momentarily clearing speed and duplex bits */
        bmcr = MiiRead(phy, MII_BMCR) &
            ~(BMCR_SPEED_100 | BMCR_DUPLEX_MODE);

        if (connect & PHY_CONN_AUTONEG) {
            /*
             * Auto-Negotiation connection mode.  Walk the entire
             * an_lookup[] table, latching all matching <phy_link> types
             * into the PHY's Auto-Negotiation Advertisement Register
             * (ANAR).  Set ANAR and BMCR when the table is exhausted.
             */
            bmcr |= (BMCR_AUTONEG_ENABLE | BMCR_RESTART_AUTONEG);
            anar = MiiRead(phy, MII_ANAR) &
                (ANAR_NEXT_PAGE | /* ANAR_ACKNOWLEDGE | */
                 ANAR_PROTOCOL_SELECTOR);
            if (connect & PHY_CONN_AUTONEG_REMOTE_FAULT)
                anar |= ANAR_REMOTE_FAULT;
            if (connect & PHY_CONN_AUTONEG_FDX_PAUSE)
                anar |= ANAR_FDX_PAUSE;
            for (lookup = &an_lookup[0]; lookup->phy_link != 0; lookup++) {
                if (connect & lookup->phy_link & PHY_CONN_TECHNOLOGY)
                    anar |= lookup->anar;
            }
            MiiWrite(phy, MII_ANAR, anar);
            phy->anar = anar;       /* cached for PhyCheck() */
            /*
             * Side-effect:  some PHYs (well, at least one vendor's)
             * require the BMCR_AUTONEG_ENABLE bit (reg 0.12) be cleared
             * and then set in the case where the device is coming up out
             * of reset or isolated state, otherwise, Auto-Negotiation
             * won't be initiated.  We'll temporarily clear the A-N bits,
             * which will then be immediately be set, below.
             */
            MiiWrite(phy, MII_BMCR,
                     (WORD) (bmcr & ~(BMCR_AUTONEG_ENABLE |
                                      BMCR_RESTART_AUTONEG)));

        } else {
            /*
             * Forced connect mode.  Walk the an_lookup[] table until a
             * matching <phy_link> type is found.  Set BMCR accordingly
             * and terminate the loop.  Doesn't yet support a dual -T2/-TX
             * device (R7145 is TBD), which likely requires an additional
             * control to select the desired 100Base-T PMA technology.
             */
            bmcr &= ~(BMCR_AUTONEG_ENABLE | BMCR_RESTART_AUTONEG);
            for (lookup = &an_lookup[0]; lookup->phy_link != 0; lookup++) {
                if (connect & lookup->phy_link & PHY_CONN_TECHNOLOGY) {
                    connect = lookup->phy_link & PHY_CONN_TECHNOLOGY;
                    if (lookup->phy_link & PHY_LINK_FDX)
                        bmcr |= BMCR_DUPLEX_MODE;
                    if (lookup->phy_link & PHY_LINK_100)
                        bmcr |= BMCR_SPEED_100;
                    break;
                }
            }
        }

        /* Issue the MII BMCR hardware command */
        MiiWrite(phy, MII_BMCR, bmcr);

        /* National PHY bug:  kick (read) PHY to restart A-N (?) */
        if ((connect & PHY_CONN_AUTONEG) && PHYID_NSC_DP83840(phy)) {
            (void) MiiRead(phy, MII_BMCR);
        }
    }

    /*
     * (4) Select default MII/HPNA ports to activate/enable and
     *     deactivate/disable.  If Auto-Negotiating both HomePNA
     *     and MII ports, activate the HomePNA port unless MII
     *     priority is specified.  If forcing or Auto-Negotiating
     *     a single port, then activate that port.
     */
    phy->active_port = phy->hpna_port;
    phy->inactive_port = phy->mii_port;
    if ((connect & PHY_CONN_MII_TECHNOLOGY) &&
            (connect & PHY_CONN_HPNA_TECHNOLOGY)) {
        /* Both ports active */
        if (connect & PHY_CONN_AUTONEG) {
            if (phy->hw_options & PHY_HWOPTS_MII_PORT_PRIORITY) {
                /* Auto-Negotiating and MII priority option */
                phy->active_port = phy->mii_port;
                phy->inactive_port = phy->hpna_port;
            }

        } else {
            /* Forced connections: disable lower priority port */
            if (phy->hw_options & PHY_HWOPTS_HPNA_PORT_PRIORITY) {
                connect &= ~PHY_CONN_MII_TECHNOLOGY;
            } else {
                connect &= ~PHY_CONN_HPNA_TECHNOLOGY;
                phy->active_port = phy->mii_port;
                phy->inactive_port = phy->hpna_port;
            }
        }

    } else if (connect & PHY_CONN_MII_TECHNOLOGY) {
        phy->active_port = phy->mii_port;
        phy->inactive_port = phy->hpna_port;
    }

    /* (5) Shave unused, residual <connect> settings and save */
    connect &= (PHY_CONN_AUTONEG              |
                PHY_CONN_AUTONEG_REMOTE_FAULT |
                PHY_CONN_AUTONEG_FDX_PAUSE    |
                PHY_CONN_TECHNOLOGY);
    if (!(connect & PHY_CONN_AUTONEG) ||
            !(connect & PHY_CONN_MII_TECHNOLOGY)) {
        connect &= ~(PHY_CONN_AUTONEG_REMOTE_FAULT |
                     PHY_CONN_AUTONEG_FDX_PAUSE);
    }
    phy->media.connect = connect;

    phy->media.cum_link_state      =
        phy->media.hpna_link_state =
        phy->media.mii_link_state  = 0;

    /*
     * (6) Activate/deactivate selected ports whenever called from
     *     PhyInit().  Thereafter, if the PHY_HWOPTS_MAC_PORT_MGMT
     *     option is not enabled, any such port changes will have to
     *     be manually performed by the caller.
     */
    if ((!IS_PHY_INITIALIZED(phy) ||
         (phy->hw_options & PHY_HWOPTS_MAC_PORT_MGMT)) &&
        !(phy->hw_options & PHY_HWOPTS_ISOLATE)) {
        PhyConfigurePort(phy, phy->media.cum_link_state);
    }
    
    return (PHY_STATUS_OK);
}

#define MAC_IO_READ(offset)         MiiIoRead(phy, (offset))
#define MAC_IO_WRITE(offset, val)   MiiIoWrite(phy, (offset), (val))

/*
 *  Register 5 (offset 0x28) Status Register One (SR1)
 */
#define MAC_SR1_REG     0x28
#define SR1_TPS         0x00000002L     /* bit[1] Tx-Process-Stopped */
#define SR1_RPS         0x00000100L     /* bit[8] Rx-Process-Stopped */
#define SR1_RS          0x000E0000L     /* bits[19:17] Rx-State */
#define SR1_RS0         0x00000000L     /*   RS0 Stop */
#define SR1_RS1         0x00020000L     /*   RS1 Reading descriptor */
#define SR1_RS2         0x00040000L     /*   RS2 Prefetch next desc. */
#define SR1_RS3         0x00060000L     /*   RS3 Waiting to receive packet */
#define SR1_RS4         0x00080000L     /*   RS4 Suspended (desc. unavail) */
#define SR1_RS5         0x000A0000L     /*   RS5 Writing desc. status */
#define SR1_RS6         0x000C0000L     /*   RS6 Flushing RxFIFO */
#define SR1_RS7         0x000E0000L     /*   RS7 DMAing RxFIFO to host mem */

#define SR1_TS          0x00700000L     /* bits[22:20] Tx-State */
#define SR1_TS0         0x00000000L     /*   TS0 Stop */
#define SR1_TS1         0x00100000L     /*   TS1 Reading descriptor */
#define SR1_TS2         0x00200000L     /*   TS2 Transmitting */
#define SR1_TS3         0x00300000L     /*   TS3 DMAing host mem to TxFIFO */
#define SR1_TS4         0x00400000L     /*   TS4 Reserved */
#define SR1_TS5         0x00500000L     /*   TS5 Processing setup packet */
#define SR1_TS6         0x00600000L     /*   TS6 Suspended (desc. unavail) */
#define SR1_TS7         0x00700000L     /*   TS7 Writing desc. status */
#define SR1_TX_STOPPED(s)               \
    (((s) & SR1_TPS)             ||     \
     (((s) & SR1_TS) == SR1_TS0) ||     \
     (((s) & SR1_TS) == SR1_TS6))
#define SR1_RX_STOPPED(s)               \
    (((s) & SR1_RPS)             ||     \
     (((s) & SR1_RS) == SR1_RS0) ||     \
     (((s) & SR1_RS) == SR1_RS3) ||     \
     (((s) & SR1_RS) == SR1_RS4))


/*
 *  Register 6 (offset 0x30) Network Access Register (NAR)
 */
#define MAC_NAR_REG     0x30
#define NAR_RX          0x00000002L     /* bit[1] */
#define NAR_FDX         0x00000200L     /* bit[9] */
#define NAR_TX          0x00002000L     /* bit[13] */
#define NAR_SPEED_10    0x00400000L     /* bit[22] */
#define NAR_PORT_MASK   0x00060000L     /* bit[18:17] */
#define NAR_PORT_HPNA   0x00020000L     /* bit[18:17] = [x:1] */
#define NAR_PORT_7WS    0x00000000L     /* bit[18:17] = [0:0] */
#define NAR_PORT_MII    0x00040000L     /* bit[18:17] = [1:0] */
#define NAR_HBD         0x00080000L     /* bit[19] */
#define NAR_SETTINGS                    \
    (NAR_FDX | NAR_SPEED_10 | NAR_PORT_MASK | NAR_HBD)

/***********************************************************************
 * PhyConfigurePort() - Configure PHY (and optional MAC) port settings
 **********************************************************************/
PHY_STATUS
PhyConfigurePort(PHY *phy, DWORD link_state)
{
    DWORD       sr1;
    DWORD       nar;
    DWORD       soft_nar;       /* may not jibe with <nar> */
    DWORD       delay10us_count;

    /*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*
     * This RSS 116xx device-dependent code rightfully belongs in the
     * portable "mii.c" module.  Oh well ...
     *!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

    if ((link_state & PHY_LINK_NEW_MEDIA) ||
            (phy->active_port != phy->enabled_port)) {

        /*
         * Upon new media and/or port change ...
         *
         * (1) Optionally stop MAC Tx/Rx, delay up to 5 maximum packet
         *     times (5 x 1518 bytes = 60,720 bits, yielding respective
         *     100/10/1 Mbit/s delays of 607.2uS/6.072mS/60.72mS)
         *     for the current transmit and/or receive processing to
         *     complete, and then disable the soon-to-be-active PHY if
         *     we're changing (MAC and PHY) ports.
         */
        if (phy->hw_options & PHY_HWOPTS_MAC_PORT_MGMT) {
            if (phy->soft_nar != NULL) {
                soft_nar = *phy->soft_nar;
                *phy->soft_nar &= ~(NAR_TX | NAR_RX);
            }
            nar = MAC_IO_READ(MAC_NAR_REG);
            if (nar & (NAR_TX | NAR_RX)) {
                MAC_IO_WRITE(MAC_NAR_REG, nar & ~(NAR_TX | NAR_RX));
                if (!(nar & NAR_SPEED_10)) {
                    delay10us_count = 61;       /* 607.2uS @ 100Mbps */
                } else if ((nar & NAR_PORT_MASK) == NAR_PORT_MII) {
                    delay10us_count = 608;      /* 6.072mS @ 10Mbps */
                } else {
                    delay10us_count = 6072;     /* 60.72ms @ 1Mbps */
                }
                do {
                    MICROSECOND_WAIT(10);
                    sr1 = MAC_IO_READ(MAC_SR1_REG);
                    if (SR1_TX_STOPPED(sr1) && SR1_RX_STOPPED(sr1)) {
                        /* Wait a few uS and reconfirm stoppage */
                        MICROSECOND_WAIT(5);
                        sr1 = MAC_IO_READ(MAC_SR1_REG);
                        if (SR1_TX_STOPPED(sr1) && SR1_RX_STOPPED(sr1)) {
                            break;
                        }
                    }

                    /* !!!!! Debug-only hook !!!!! */
                    if (delay10us_count == 1) {
                        break;  /* Strikeout: set breakpoint here */
                    }
                    /* !!!!! Debug-only hook !!!!! */

                } while (--delay10us_count > 0);
            }

            /*
             * Temporarily tri-state the active, soon-to-be-enabled port
             * only upon initial or subsequent port changes.
             */
            if (phy->active_port != phy->enabled_port ) {
                (void) PhyDisable(phy, phy->active_port);
            }
        }

        /* (2) Disable the inactive PHY */
        (void) PhyDisable(phy, phy->inactive_port);
        phy->disabled_port = phy->inactive_port;

        /* (3) Optionally reconfigure the MAC's NAR_SETTINGS */
        if (phy->hw_options & PHY_HWOPTS_MAC_PORT_MGMT) {
            nar &= ~NAR_SETTINGS;
            nar |= NAR_HBD;     /* SQE heartbeat disabled MII/HomePNA */
            if (phy->active_port & PHY_PORT_HPNA_INT) {
                nar |= (NAR_PORT_HPNA | NAR_SPEED_10);  /* internal */

            } else if (phy->active_port & PHY_PORT_HPNA_EXT) {
                nar |= (NAR_PORT_7WS | NAR_SPEED_10);   /* external */

            } else {
                nar |= NAR_PORT_MII;    /* inferred */
                if ((link_state & PHY_LINK_UP) ||
                        ((link_state = phy->media.mii_link_state) != 0)) {
                    /* Go with current or previous MII link settings */
                    if (link_state & PHY_LINK_10) {
                        nar |= NAR_SPEED_10;    /* [sic] */
                    }
                    if (link_state & PHY_LINK_FDX) {
                        nar |= NAR_FDX;
                    }

                } else {
                    /* MII link never established: go with connect */
                    link_state = phy->media.connect;
                    if (link_state & (PHY_LINK_100BASE_T2FD |
                                      PHY_LINK_100BASE_TXFD)) {
                        if (!(link_state & PHY_LINK_AUTONEG)) {
                            /* A-N @ half-duplex, forced @ full-duplex */
                            nar |= NAR_FDX;
                        }
                    } else if (link_state & (PHY_LINK_100BASE_T2 |
                                             PHY_LINK_100BASE_TX |
                                             PHY_LINK_100BASE_T4)) {
                        ;       /* nada */

                    } else if (link_state & PHY_LINK_10BASE_TFD) {
                        nar |= NAR_SPEED_10;
                        if (!(link_state & PHY_LINK_AUTONEG)) {
                            /* A-N @ half-duplex, forced @ full-duplex */
                            nar |= NAR_FDX;
                        }
                    } else {
                        nar |= NAR_SPEED_10;
                    }
                }
            }
            /* Update soft and hard NAR, but don't yet enable TX/RX */
            if (phy->soft_nar != NULL) {
                soft_nar &= ~NAR_SETTINGS;
                soft_nar |= (nar & NAR_SETTINGS);
                *phy->soft_nar = soft_nar & ~(NAR_TX | NAR_RX);
            }
            MAC_IO_WRITE(MAC_NAR_REG, nar & ~(NAR_TX | NAR_RX));
        }
        /* (4) Enable the active PHY */
        (void) PhyEnable(phy, phy->active_port);
        phy->enabled_port = phy->active_port;

        /* (5) Optionally restart the MAC Tx/Rx */
        if (phy->hw_options & PHY_HWOPTS_MAC_PORT_MGMT) {
            if (phy->soft_nar != NULL) {
                *phy->soft_nar = soft_nar;
            }
            MAC_IO_WRITE(MAC_NAR_REG, nar);
        }
    }

    return (PHY_STATUS_OK);
}

/***********************************************************************
 * PhyCapabilityGet() - Get PHY technology capabilities
 **********************************************************************/
DWORD
PhyCapabilityGet(PHY *phy)
{
    return (phy->media.capability);
}

/***********************************************************************
 * PhyCheck/PhyCheck2() - Assess link status, speed and duplex mode
 **********************************************************************/
DWORD
PhyCheck(PHY *phy)
{
    return (PhyCheck2(phy, FALSE));
}

DWORD
PhyCheck2(PHY *phy, BOOLEAN force_hpna_heartbeat)
{
    DWORD           connect          = phy->media.connect;
    DWORD           hpna_link_state  = phy->media.hpna_link_state;
    DWORD           cum_link_state   = phy->media.cum_link_state;
    DWORD           mii_link_state   = phy->media.mii_link_state;
    DWORD           mii_link_error   = 0;
    WORD            bmsr;
    WORD            aner;
    WORD            negotiation;
    AN_LOOKUP      *table;
    BYTE            istat_lsb;
    PHY_PORT        port_select;

    if (!IS_PHY_INITIALIZED(phy)) {
        return (PHY_LINK_INIT_ERROR);
    }


    /*----------------------------------------------------------------*
     * (1) Examine HomePNA PHY
     *----------------------------------------------------------------*/

    if (!(connect & PHY_CONN_HPNA_TECHNOLOGY)) {
        istat_lsb = 0;  /* affects HomePNA heartbeat, Item (6) below */

    } else if (phy->hpna_isolated) {
        istat_lsb = 0;  /* ignore HPNA status when isolated */

    } else {
        /* Read and clear ISTAT[0].Bit[3:2] => PHY-level [Rx:Tx] */
        istat_lsb = SpiReadByte(phy, HPNA_ISTAT);
        if ((istat_lsb & (HPNA_ISTAT_PACKET_TRANSMITTED |
                          HPNA_ISTAT_PACKET_RECEIVED)) != 0x00) {
            SpiWriteByte(phy, HPNA_ISTAT, 0x00);
        }

        /* PHY Rx: assume link up, else going down (pending) or down */
        if (istat_lsb & HPNA_ISTAT_PACKET_RECEIVED) {
            phy->hpna_link_pending = 0;         /* FALSE */
            if (!(hpna_link_state & PHY_LINK_UP)) {
                /* Latch 1st and subsequent PHY_LINK_NEW_MEDIA change */
                hpna_link_state =
                    (PHY_LINK_UP | PHY_LINK_1  |
                     (connect & PHY_CONN_HPNA_TECHNOLOGY));
                if ((hpna_link_state ^ phy->media.hpna_link_state) &
                        (PHY_LINK_SPEED | PHY_LINK_FDX |
                         PHY_LINK_HPNA_TECHNOLOGY)) {
                    phy->media.hpna_link_state = hpna_link_state;
                    hpna_link_state |= PHY_LINK_NEW_MEDIA;
                } else {
                    phy->media.hpna_link_state = hpna_link_state;
                }
            }

        } else if (hpna_link_state & PHY_LINK_UP) {
            if (((phy->hw_options & PHY_HWOPTS_ALT_1HZ_HPNA_LINK) &&
                 !(istat_lsb & HPNA_ISTAT_PACKET_TRANSMITTED) &&
                 (++phy->hpna_link_pending >= 4)) ||

               (!(phy->hw_options & PHY_HWOPTS_ALT_1HZ_HPNA_LINK) &&
                (++phy->hpna_link_pending >= 4))) {

                /* Link down: report current connect state */
                phy->media.hpna_link_state &= ~PHY_LINK_UP;
                hpna_link_state = connect & PHY_CONN_HPNA_TECHNOLOGY;
            }
        } else {
            /* Link down: report current connect state */
            phy->media.hpna_link_state &= ~PHY_LINK_UP;
            hpna_link_state = connect & PHY_CONN_HPNA_TECHNOLOGY;
        }
        
    }

    /*----------------------------------------------------------------*
     * (2) Examine MII PHY
     *----------------------------------------------------------------*/

    if( (connect & PHY_CONN_MII_TECHNOLOGY) && (phy->mii_isolated == FALSE)){
        /*
         * Important note:  must double-clutch BMSR when link status
         * is down, since the BMSR_LINK_STATUS bit is latched LOW until
         * read.  A second BMSR read ensures current link state.
         */
        bmsr = MiiRead(phy, MII_BMSR);
        if (!(bmsr & BMSR_LINK_STATUS)) {
            bmsr |= MiiRead(phy, MII_BMSR); /* |= retains error latches */

            /*
             * Upon "long-term" link down transition, disable power-saving
             * TDK RXCC option (RX_CLK is held low when no data is
             * received), as it adversely impacts Auto-Negotiation.  We'll
             * restart Auto-Negotiation as a precautionary measure.
             */
            if ((phy->hw_options & PHY_HWOPTS_CK25DIS) &&
                    PHYID_TDK_2120(phy) &&
                    (mii_link_state & PHY_LINK_UP) &&
                    !(bmsr & BMSR_LINK_STATUS)) {
                MII_CLR_BIT(phy, MII_R16, TDK_R16_RXCC);
                if (connect & PHY_CONN_AUTONEG) {
                    MII_SET_BIT(phy, MII_BMCR, BMCR_RESTART_AUTONEG);
                }
            }

            /* Link down: report current connect state */
            phy->media.mii_link_state &= ~PHY_LINK_UP;
            mii_link_state = connect & (PHY_CONN_MII_TECHNOLOGY       |
                                        PHY_CONN_AUTONEG_REMOTE_FAULT |
                                        PHY_CONN_AUTONEG_FDX_PAUSE);

        }

        /*
         * Re-evaluate link configuration for down-to-up transitions,
         * otherwise report present (connecting or operational) link state.
         */
        if ((bmsr & BMSR_LINK_STATUS) && !(mii_link_state & PHY_LINK_UP)) {
            mii_link_state = 0;
            if (!(connect & PHY_CONN_AUTONEG)) {
                /*
                 * Forced connection mode:  assume the forced state,
                 * walking the an_lookup[] table to pick up associated
                 * PHY_LINK_FDX and PHY_LINK_100/10 status bits.
                 */
                for (table = &an_lookup[0]; table->phy_link != 0; table++) {
                    if (connect & table->phy_link & PHY_LINK_TECHNOLOGY) {
                        mii_link_state = table->phy_link | PHY_LINK_UP;
                        break;
                    }
                }

            } else {
                /*
                 * Auto-Negotiation mode:  infer link speed/mode based
                 * on Link Control Words (LCWs) advertised and received
                 * from our link partner.  Simply walk the negotiated
                 * settings against the an_lookup[] table, ordered from
                 * highest to lowest Auto-Negotiation priority, reporting
                 * first (highest) matching capability.  Special case:
                 * the Davicom PHY doesn't cleanse (zero) the ANLPAR link
                 * partner register when transitioning from an A-N partner
                 * to a non A-N partner, so check its ANER expansion
                 * register for a new link page notification.  Not all PHYs
                 * indicate new page in all modes, e.g., the QSI 970 and
                 * Level One (LXT 901) PHYs' "ANLPAR spoofing" won't
                 * indicate ANER new page when linked to a non A-N partner,
                 * and the pre-RevA NSC 83840 apparently doesn't indicate
                 * next page even when linked with an A-N partner.
                 */
                if ((bmsr & BMSR_AUTONEG_COMPLETE) &&
                        (((aner = MiiRead(phy, MII_ANER)) &
                           ANER_PAGE_RECEIVED) ||
                         !PHYID_DAVICOM_9101(phy))) {
                    negotiation = (phy->anar & MiiRead(phy, MII_ANLPAR));
                    for (table = &an_lookup[0]; table->anar != 0; table++) {
                        if (negotiation & table->anar) {
                            mii_link_state = (PHY_LINK_UP      |
                                              PHY_LINK_AUTONEG |
                                              table->phy_link);
                            break;
                        }
                    }
                    if (aner & ANER_PARALLEL_DETECTION_FAULT) {
                        mii_link_error |= PHY_LINK_PARALLEL_ERROR;
                    }
                }
                if (PHYID_KENDIN_KS8737(phy)) {
                    /*
                     * Though the Kendin PHY supports "ANLPAR spoofing,"
                     * subsequent to link with Auto-Negotiating peer,
                     * it fails to update its ANLPAR register, resulting
                     * in false mode selection.  We'll correct via
                     * vendor-specific register 31.
                     */
                    WORD    r31 = MiiRead(phy, MII_R31);
                    if (!(r31 & MII_BIT4)) {
                        /*
                         * Not full-duplex, which can only result from
                         * a true Auto-Negotiation handshake.
                         */
                        mii_link_state = (PHY_LINK_UP | PHY_LINK_AUTONEG);
                        if (r31 & MII_BIT2) {
                            mii_link_state |=
                                (PHY_LINK_10  | PHY_LINK_10BASE_T);
                        } else {
                            mii_link_state |=
                                (PHY_LINK_100 | PHY_LINK_100BASE_TX);
                        }
                    }

                } else if (mii_link_state == 0) {
                    /*
                     * Parallel Detection: link but no Auto-Negotiation
                     * match, which implies our link partner is not A-N
                     * capable.  Determine link technology via PHY
                     * vendor-specific status registers.
                     */
                    if (PHYID_NSC_DP83840(phy)) {
                        mii_link_state = PHY_LINK_UP | PHY_LINK_AUTONEG;
                        if (MiiRead(phy, MII_PAR) & PAR_SPEED_10) {
                            mii_link_state |=
                                (PHY_LINK_10  | PHY_LINK_10BASE_T);
                        } else {
                            mii_link_state |=
                                (PHY_LINK_100 | PHY_LINK_100BASE_TX);
                        }

                    } else if (PHYID_GEC_NWK936(phy)) {
                        /*
                         * The GEC PHY apparently sets BMCR speed/duplex
                         * bits based on the A-N outcome.  Though their spec
                         * says to use extended register 25 for current
                         * speed and duplex state, we'll stick to the
                         * standard BMCR register, although most PHYs
                         * don't reflect current speed/duplex state in
                         * the BMCR when Auto-Negotiating....
                         */
                        mii_link_state = PHY_LINK_UP | PHY_LINK_AUTONEG;
                        if (MiiRead(phy, MII_BMCR) & BMCR_SPEED_100) {
                            mii_link_state |=
                                (PHY_LINK_100 | PHY_LINK_100BASE_TX);
                        } else {
                            mii_link_state |=
                                (PHY_LINK_10  | PHY_LINK_10BASE_T);
                        }

                    } else if (PHYID_TDK_2120(phy)) {
                        /*
                         * The TDK PHY's diagnostic register 18 (decimal)
                         * MR18.RATE, bit-10 (0x400) indicates the result of
                         * Auto-Negotiation data rate arbitration.  However,
                         * when using their pre-production Rev 3 device,
                         * our testing apparently revealed bit-9 (0x200) as
                         * the RATE indicator when connecting to dumb,
                         * non Auto-Negotiating hubs at 10 or 100 Mbit/sec,
                         * half-duplex.  TDK support confirms that bit-10
                         * RATE is a new feature in the Rev 7 device.  As
                         * such, we'll go with this apparent bit-9 heuristic
                         * on the earlier, Rev-3 device.
                         */
                        mii_link_state = PHY_LINK_UP | PHY_LINK_AUTONEG;
                        if (MiiRead(phy, MII_R18) &
                                ((PHYID_REV(phy) >= 7) ? TDK_R18_RATE_100
                                                       : MII_BIT9)) {
                            mii_link_state |=
                                (PHY_LINK_100 | PHY_LINK_100BASE_TX);
                        } else {
                            mii_link_state |=
                               (PHY_LINK_10   | PHY_LINK_10BASE_T);
                        }

                    } else if (PHYID_DAVICOM_9101(phy)) {
                        /*
                         * The Davicom PHY apparently (as of 21Jan98 we
                         * have no data sheet) identifies auto-negotiation
                         * results in bits 12-15 of register 17 (decimal).
                         * 8nnn = 100-FDX, 4nnn = 100-HDX, 2nnn = 10-FDX,
                         * and 1nnn = 10-HDX.  Since an A-N partner's
                         * advertisement should be correctly reflected in
                         * the ANLPAR (register 5), we'll only troll for
                         * dumb, non A-N partners' half-duplex results here.
                         */
                        mii_link_state = PHY_LINK_UP | PHY_LINK_AUTONEG;
                        switch (MiiRead(phy, MII_R17) & 0xF000) {
                            default:
                                mii_link_state = 0;
                                break;
                            case 0x4000:
                                mii_link_state |=
                                    (PHY_LINK_100 | PHY_LINK_100BASE_TX);
                                break;
                            case 0x1000:
                                mii_link_state |=
                                    (PHY_LINK_10  | PHY_LINK_10BASE_T);
                                break;
                        }
                    } else if (PHYID_LXT_971(phy)) {

                        // Unspecified, based on lab test observation 03Dec99                        */
                        mii_link_state = PHY_LINK_UP | PHY_LINK_AUTONEG;
                        if(MiiRead(phy, MII_R17) & 0x4000) {
                            mii_link_state |= (PHY_LINK_100 | PHY_LINK_100BASE_TX);
                        }
                        else {
                            mii_link_state |= (PHY_LINK_10  | PHY_LINK_10BASE_T);
                        }
                    }
                } /* end if - SW parallel detection logic */
            } /* end if - Auto-Negotiating */

            /* Latch 1st and subsequent PHY_LINK_NEW_MEDIA changes */
            if (mii_link_state == 0) {
                /* ??? Couldn't resolve link; need PHY_LINK_UNKNOWN err */
                mii_link_state = connect & (PHY_CONN_MII_TECHNOLOGY       |
                                            PHY_CONN_AUTONEG_REMOTE_FAULT |
                                            PHY_CONN_AUTONEG_FDX_PAUSE);
            } else if ((mii_link_state ^ phy->media.mii_link_state) &
                     (PHY_LINK_SPEED | PHY_LINK_FDX |
                      PHY_LINK_MII_TECHNOLOGY)) {
                phy->media.mii_link_state = mii_link_state;
                mii_link_state |= PHY_LINK_NEW_MEDIA;
            } else {
                phy->media.mii_link_state = mii_link_state;
            }

            /* Enable TDK RXCC option only on link down-to-up @ 100 Mbit */
            if ((phy->hw_options & PHY_HWOPTS_CK25DIS) &&
                    PHYID_TDK_2120(phy)) {
                if (mii_link_state & PHY_LINK_100) {
                    MII_SET_BIT(phy, MII_R16, TDK_R16_RXCC);
                } else {
                    MII_CLR_BIT(phy, MII_R16, TDK_R16_RXCC);
                }
            }
        } /* endif - MII link down-to-up transition */

        /* Affix newly detected errors latched by PHY hardware */
        if ((bmsr & BMSR_JABBER_DETECT) &&
                (mii_link_state & (PHY_LINK_10BASE_TFD |
                                   PHY_LINK_10BASE_T))) {
            mii_link_error |= PHY_LINK_JABBER;
        }
        if (bmsr & BMSR_REMOTE_FAULT) {
            mii_link_error |= PHY_LINK_REMOTE_FAULT;
        }
    }
    /*----------------------------------------------------------------*
     * (3) Form cumulative link state based on HomePNA and MII states
     *----------------------------------------------------------------*/

    /* Initial state indicates Auto-Negotiation directive */
    cum_link_state = (connect & PHY_LINK_AUTONEG);

    if (!(connect & PHY_CONN_HPNA_TECHNOLOGY)) {
        /*
         * (3a) No HomePNA connect, so infer MII-only operation
         */
        cum_link_state |= mii_link_state;
        port_select = phy->mii_port;

    } else if (!(connect & PHY_CONN_MII_TECHNOLOGY)) {
        /*
         * (3b) No MII connect, so infer HomePNA-only operation
         */
        cum_link_state |= hpna_link_state;
        port_select = phy->hpna_port;

        /*
         * (3c) Simultaneous HomePNA and MII operation ...
         */
    } else if (!((hpna_link_state | mii_link_state) & PHY_LINK_UP)) {

        /*
         * (3c-1) Neither device reporting link: choose a port.
         *
         * By default, retain the current, active port.  After
         * LINK_DOWN_HOP_COUNT (nominally 5) successive cycles of
         * link down, we can optionally switch (hop) to the other,
         * inactive port or lock-in on the MII or HomePNA port if
         * the respective PHY_HWOPTS_PORT_HOPPING, PHY_HWOPTS_MII_ 
         * PORT_PRIORITY or PHY_HWOPTS_HPNA_PORT_PRIORITY options
         * are in effect.  Special port-hopping note:  certain
         * tested MII PHYs such as the TDK, QSI and GEC don't require
         * hopping to achieve link, thus we won't hop thereupon
         * to improve our odds of achieving link on the HomePNA
         * device.
         */
        cum_link_state |= (hpna_link_state | mii_link_state);
        port_select = phy->active_port;
        if (++phy->link_down_count >= LINK_DOWN_HOP_COUNT) {
            phy->link_down_count = 0;
            if ((phy->hw_options & PHY_HWOPTS_PORT_HOPPING) &&
                    (phy->inactive_port != PHY_PORT_NONE)) {
                port_select = phy->inactive_port;   /* (hop) */
            } else if (phy->hw_options & PHY_HWOPTS_MII_PORT_PRIORITY) {
                port_select = phy->mii_port;
            } else if (phy->hw_options & PHY_HWOPTS_HPNA_PORT_PRIORITY) {
                port_select = phy->hpna_port;
            }
        }

    } else {
        /*
         * (3c-2) One or both devices reporting link: choose a port.
         *
         * If both devices report link, retain the active port if
         * previously linked.  Otherwise, default to the MII link
         * status and select the MII port, since 10/100 Mbit/s is
         * considered a greater capability than 1 Mbit/s HomePNA.
         * This can be overriden via the PHY_HWOPTS_HPNA_PORT_PRIORITY
         * option.
         *
         * If only one device is reporting link, select that device's
         * port and use its associated link status.
         */
        phy->link_down_count = 0;
        if (hpna_link_state & mii_link_state & PHY_LINK_UP) {
            /* Both devices reporting link: choose a winner */
            if (phy->media.cum_link_state & PHY_LINK_UP) {
                port_select = phy->active_port;
                if (port_select == phy->hpna_port) {
                    cum_link_state |= hpna_link_state;
                } else {
                    cum_link_state |= mii_link_state;
                }

            } else if (phy->hw_options &
                       PHY_HWOPTS_HPNA_PORT_PRIORITY) {
                cum_link_state |= hpna_link_state;
                port_select = phy->hpna_port;

            } else {
                /* 10/100 Mbit/s MII > 1-Mit/s HomePNA */
                cum_link_state |= mii_link_state;
                port_select = phy->mii_port;
            }

        } else if (mii_link_state & PHY_LINK_UP) {
            /* MII-only link up */
            cum_link_state |= mii_link_state;
            port_select = phy->mii_port;

        } else {
            /* HomePNA-only link up */
            cum_link_state |= hpna_link_state;
            port_select = phy->hpna_port;
        }
    }

    /*----------------------------------------------------------------*
     * (4) Select port and merge into cumulative link state
     *----------------------------------------------------------------*/

    switch (port_select) {
        case PHY_PORT_HPNA_INT:  cum_link_state |= PHY_LINK_PORT_HPNA; break;
        case PHY_PORT_HPNA_EXT:  cum_link_state |= PHY_LINK_PORT_7WS;  break;
        case PHY_PORT_MII:       cum_link_state |= PHY_LINK_PORT_MII;  break;
        default:                 /* ??? */                             break;
    }
    if (!(cum_link_state & PHY_LINK_UP)) {
        phy->media.cum_link_state &= ~PHY_LINK_UP;
    } else {
        if ((cum_link_state ^ phy->media.cum_link_state) &
                (PHY_LINK_PORT | PHY_LINK_FDX | PHY_LINK_SPEED |
                 PHY_LINK_TECHNOLOGY)) {
            cum_link_state |= PHY_LINK_NEW_MEDIA;
        }
        phy->media.cum_link_state = cum_link_state & ~PHY_LINK_NEW_MEDIA;

        /* Suppress MII errors when the HomePNA link is up and active */
        if (cum_link_state & PHY_LINK_HPNA_TECHNOLOGY) {
            mii_link_error = 0;
        }
    }

    /*----------------------------------------------------------------*
     * (5) Optionally reconfigure upon new-media and/or port switch
     *----------------------------------------------------------------*/
    if ((cum_link_state & PHY_LINK_NEW_MEDIA) ||
            ((port_select != phy->active_port) &&
             (port_select != PHY_PORT_NONE))) {
        if ((port_select != phy->active_port) &&
                (port_select != PHY_PORT_NONE)) {
            phy->inactive_port = phy->active_port;
            phy->active_port = port_select;
        }
        if ((phy->hw_options & PHY_HWOPTS_MAC_PORT_MGMT) &&
                !(phy->hw_options & PHY_HWOPTS_ISOLATE)) {
            PhyConfigurePort(phy, cum_link_state);
        }
    }

    /*----------------------------------------------------------------*
     * (6) PHY Tx: send a HomePNA heartbeat if forcing heartbeats,
     *     or nothing transmitted, or PHY_HWOPTS_ALT_1HZ_HPNA_LINK
     *     and nothing transmitted and nothing received.
     *----------------------------------------------------------------*/
    if ((force_hpna_heartbeat ||
            (!(istat_lsb & HPNA_ISTAT_PACKET_TRANSMITTED) &&
             (!(istat_lsb & HPNA_ISTAT_PACKET_RECEIVED) ||
              !(phy->hw_options & PHY_HWOPTS_ALT_1HZ_HPNA_LINK)))) &&
            phy->callback.enabled &&
            (phy->enabled_port == phy->hpna_port) &&
            (phy->hpna_port != PHY_PORT_NONE)) {
        /*
         * No Tx, so send HPNA-link pulse message to other nodes.
         * Use null-addressed LLC Test frame to avoid soliciting
         * MAC-level responses.
         */
        if (phy->callback.one_shot) {
            phy->callback.enabled = FALSE;
        }
        (*phy->callback.mac_tx_function)(phy->callback.mac_adapter,
                                         &phy->callback.test_frame[0],
                                         phy->callback.test_frame_size);
    }

    return (cum_link_state | mii_link_error);
}

/***********************************************************************
 * PhyLastLinkUp() - Return current or last known, valid link state
 **********************************************************************/
DWORD
PhyLastLinkUp(PHY *phy)
{
    if (phy->media.cum_link_state != 0) {
        return (phy->media.cum_link_state | PHY_LINK_UP);
    }
    return (0);         /* valid link never established */
}

/***********************************************************************
 * PhyShutdown() - Disable (isolate) and power-down the PHY
 **********************************************************************/
PHY_STATUS
PhyShutdown(PHY *phy)
{
    (void) PhyDisable(phy, PHY_PORT_MII_DEVICE);
    MII_SET_MASK(phy, MII_BMCR, BMCR_POWERDOWN);

    return (PHY_STATUS_OK);
}

/***********************************************************************
 * PhyLoopback() - Enable/disable PHY loopback (one of four modes)
 **********************************************************************/
PHY_STATUS
PhyLoopback(PHY *phy, PHY_LOOPBACK mode)
{
    WORD        lbremr;

    if (!IS_PHY_INITIALIZED(phy)) {
        return (PHY_STATUS_UNINITIALIZED);
    }

    switch (mode) {
        default:
            mode = PHY_LOOPBACK_OFF;
            /* fall through ... */
        case PHY_LOOPBACK_MII:       /* MAC loopback at MII Tx/Rx */
            lbremr = LBREMR_LB_OFF;
            break;

        case PHY_LOOPBACK_PMD:       /* Loopback at the PMD Tx/Rx */
            lbremr = LBREMR_LB_XCVR;
            break;

        case PHY_LOOPBACK_REMOTE:    /* Media Rx Data looped onto Tx at
                                      * MII (e.g., link BER testing) */
            lbremr = LBREMR_LB_REMOTE;
            break;

        case PHY_LOOPBACK_10BASE_T:
            lbremr = LBREMR_LB_10BT;
            break;
    }

    /* Configure BMCR and LBREMR registers' loopback control bits */
    phy->loopback = mode;
    if (mode == PHY_LOOPBACK_MII) {
        MII_SET_BIT(phy, MII_BMCR, BMCR_LOOPBACK);
    } else {
        MII_CLR_BIT(phy, MII_BMCR, BMCR_LOOPBACK);
    }
    MII_CLR_MASK(phy, MII_LBREMR, LBREMR_LB_MASK);
    MII_SET_MASK(phy, MII_LBREMR, lbremr);

    return (PHY_STATUS_OK);
}

/***********************************************************************
 * PhyColTest() - Enable/disable collision test (diagnostic)
 **********************************************************************/
PHY_STATUS
PhyColTest(PHY *phy, BOOLEAN enable)
{
    phy = phy;
    return ((enable) ? PHY_STATUS_UNSUPPORTED_FUNCTION : PHY_STATUS_OK);
}

/***********************************************************************
 * PhyIdGet() - Read HW ID from registers 2 and 3
 **********************************************************************/
DWORD
PhyIdGet(PHY *phy)
{
    return (phy->id);
}

/***********************************************************************
 * PhyCallbackEnable() - Register a MAC transmit routine for HomePNA use
 **********************************************************************/
PHY_CALLBACK
PhyCallbackEnable(PHY          *phy,
                  PHY_CALLBACK  callback_func,
                  void         *mac_adapter,
                  void         *mac_src_addr,
                  BOOLEAN       one_shot)
{
    static BYTE          null_mac_addr[6];      /* = {0}; */
    BYTE                *mac_addr;
    PHY_CALLBACK         previous_callback;
    int                  n;

    previous_callback             = phy->callback.mac_tx_function;
    phy->callback.mac_tx_function = callback_func;

    if (callback_func == (PHY_CALLBACK) NULL) {
        phy->callback.enabled     = FALSE;

    } else {
        phy->callback.one_shot    = one_shot;
        phy->callback.mac_adapter = mac_adapter;

        /* Self-addressed test frame (to me, from me) */
        if ((mac_addr = mac_src_addr) == NULL) {
            mac_addr = &null_mac_addr[0];
        }
        for (n = 0; n < 6; n++) {
            phy->callback.test_frame[n] = *mac_addr++;  /* destAddr */
        }
        mac_addr -= 6;
        for (n = 6; n < 12; n++) {
            phy->callback.test_frame[n] = *mac_addr++;  /* srcAddr */
        }

        phy->callback.enabled     = TRUE;
    }

    return (previous_callback);
}

/***********************************************************************
 * PhyWakeOnLink() - Enable/disable PHY's power management wake-up logic
 **********************************************************************/
PHY_STATUS
PhyWakeOnLink(PHY *phy, BOOLEAN enable)
{
    if (phy->hpna_port != PHY_PORT_NONE) {
        if (!enable) {
            SpiWriteWord(phy, HPNA_IMASK, 0);
        } else {
            if (phy->enabled_port != phy->hpna_port) {
                // ??? PhyConfigurePort(phy, PHY_LINK_NEW_MEDIA);
            }
            SpiWriteWord(phy, HPNA_IMASK, HPNA_IMASK_PACKET_RECEIVED);
            SpiWriteWord(phy, HPNA_ISTAT, 0);
        } 
    }

    return (PHY_STATUS_OK);
}

/*
 * Following routines are added for Ramses
 */
#define MAC_IO_READ(offset)         MiiIoRead(phy, (offset))
#define MAC_IO_WRITE(offset, val)   MiiIoWrite(phy, (offset), (val))

#define MAC_SPR_REG             0x48        /* same as MAC_MDIO_REG */
#define SPR_FRCS                0x00000010L /* 11623 serial Flash ROM Chip Select */
#define EEP_DI                  0x00000004L /* 11623 serial EEPROM DI */

#define MAC_GPIO_REG        0x78
#define RAMSE_ENABLE        0x10000L    /* Set LAN_GPIO[0] to Hi for Ramses enable */
#define GPIO2_HI            0x40000L    /* Set GPIO[2] to high */

#define DEFAULT_CEILING         0x7f

#define DEFAULT_FLOOR_X2           0x10
#define DEFAULT_SLC_PEAK_MID_X2    0x80

#define DEFAULT_FLOOR           0x8
#define DEFAULT_SLC_PEAK_MID    0x50
#define DEFAULT_NSE_ATTACK      0x72

#define HPNA_CONTROL_DISABLE_HALF_MODE  0x1000

static BOOLEAN
IsRamseX2 (PHY *phy)
{
    WORD wCtrl;

    if(phy->hw_options & PHY_HWOPTS_X3_DEVICE)
        return FALSE;
    else {
        /*
        * Check for Ramse X4(bit 0 in HPNA_CONTROL is 0 and read only)
        */
        wCtrl = SpiReadWord(phy, HPNA_CONTROL);
        if ((wCtrl & 1) == 0) {
            SpiWriteWord(phy, HPNA_AFE_CONTROL, (WORD)(wCtrl | 1));
            wCtrl = SpiReadWord(phy, HPNA_CONTROL);
            if ((wCtrl & 1) == 0)
                return FALSE;       /* Yes, it is X4 */
        }
    }
    return TRUE;
}

static PHY_STATUS
RamseInit (PHY *phy)
{
    PHY_STATUS  status  = PHY_STATUS_NOT_FOUND;
    WORD wAfeCtl, wCtrl;
    BYTE setFloor = DEFAULT_FLOOR, setSlcPeakMid = DEFAULT_SLC_PEAK_MID;

    if (phy->hpna_port == PHY_PORT_HPNA_EXT && phy->gpio_spi_phy_csn == 0)
    {
        if (IsRamseX2(phy)) {
            setFloor = DEFAULT_FLOOR_X2;
            setSlcPeakMid = DEFAULT_SLC_PEAK_MID_X2;
        }
        /* Stop slice level adaption */
        wCtrl = SpiReadWord(phy, HPNA_CONTROL);
        wCtrl |= HPNA_CONTROL_STOP_SLICE_ADAPT;
        SpiWriteWord(phy, HPNA_CONTROL, wCtrl);

        /* Setup for Ramses */
        SpiWriteByte(phy, HPNA_NSE_FLOOR, setFloor);
        SpiWriteByte(phy, HPNA_NSE_CEILING, DEFAULT_CEILING);
        SpiWriteByte(phy, HPNA_SLC_PEAK_MID, setSlcPeakMid);
        SpiWriteByte(phy, HPNA_NSE_ATTACK, DEFAULT_NSE_ATTACK);

        wAfeCtl = SpiReadWord(phy, HPNA_AFE_CONTROL);
        wAfeCtl &= 0xff;        /* Clean up Hi byte */
        wAfeCtl |= (HPNA_AFE_CONTROL_PEAK_DELAY | HPNA_AFE_CONTROL_NOISE_SLICE);
        SpiWriteWord(phy, HPNA_AFE_CONTROL, wAfeCtl);

        wCtrl |= HPNA_CONTROL_IGNORE_REMOTE_CMD | HPNA_CONTROL_DISABLE_HALF_MODE;

        if (phy->ramse_ctrl_reg != 0) {
            /* Restore the control register */
            SpiWriteWord(phy, HPNA_CONTROL, phy->ramse_ctrl_reg);

            /* Ramses has a bug, which is Hi/Lo power mode can't be set/reset
               by HPNA_CONTROL */
            if (phy->ramse_ctrl_reg & HPNA_CONTROL_HIGH_POWER) {
                wAfeCtl |= HPNA_CONTROL_HIGH_POWER;
                SpiWriteWord(phy, HPNA_AFE_CONTROL, wAfeCtl);
            }
        } else {
            /* Enable slice level adaption */
            wCtrl &= ~HPNA_CONTROL_STOP_SLICE_ADAPT;
            SpiWriteWord(phy, HPNA_CONTROL, wCtrl);
        }
        status = PHY_STATUS_OK;
    } 
    return (status);
}

static void
DisableRamse (PHY *phy)
{
    DWORD       gpio_data;
        DWORD       spr_data;

    if (phy->hpna_port == PHY_PORT_HPNA_EXT && phy->gpio_spi_phy_csn == 0)
    {
        /* Detected Ramses */

        /* Save the control register */
        phy->ramse_ctrl_reg = SpiReadWord(phy, HPNA_CONTROL);

        if (phy->hw_options & PHY_HWOPTS_GP0_SPI_DOUT) {
            /* FlashROM CS# = 1 (ROM_CS# = 0) */
            /* Hold in RESET# state.                                    */

            spr_data = MAC_IO_READ(MAC_SPR_REG);
            spr_data |= SPR_FRCS;
            MAC_IO_WRITE(MAC_SPR_REG, spr_data);
        } else {
            /* Here we assume the LAN_GPIP[0] has been set to output.   */
            /* Hold in RESET# state.                                    */
            gpio_data = MAC_IO_READ(MAC_GPIO_REG);
            gpio_data &= ~RAMSE_ENABLE; 
            MAC_IO_WRITE(MAC_GPIO_REG, gpio_data);
        }
    }
    return;
}

static void
EnableRamse (PHY *phy)
{
    DWORD       gpio_data;
    DWORD       spr_data;

    if (phy->hpna_port == PHY_PORT_HPNA_EXT && phy->gpio_spi_phy_csn == 0)
    {
        /* Detected Ramses */
        if (phy->hw_options & PHY_HWOPTS_GP0_SPI_DOUT) {
            /* FlashROM CS# = 0 (ROM_CS# = 1) */
            /* release RESET#                                           */
            spr_data = MAC_IO_READ(MAC_SPR_REG);
            spr_data &= ~SPR_FRCS;
            MAC_IO_WRITE(MAC_SPR_REG, spr_data);
        } else {
            /* Here we assume the LAN_GPI0[0] has been set to output.   */
            /* release RESET#                                           */
            gpio_data = MAC_IO_READ(MAC_GPIO_REG);
            gpio_data |= RAMSE_ENABLE; 
            MAC_IO_WRITE(MAC_GPIO_REG, gpio_data);
        }
        MILLISECOND_WAIT(1);    /* a la "mii.c" GPIO_SETTLE_TIME */

        SpiWriteEnable(phy);
        RamseInit(phy);
    }
    return;
}

void
RamseD3Set (PHY *phy)
{
    if (phy->hpna_port == PHY_PORT_HPNA_EXT && phy->gpio_spi_phy_csn == 0)
    {
        /* Detected Ramses */
        if (phy->hw_options & PHY_HWOPTS_GP0_SPI_DOUT) {
            DWORD       spr_data, gpio_data;

            /* The Flash ROM CS (ROM_CS# = 1)is connected to Ramses reset */
            /* pin. In D3 cold, ROM CS will be tri-state.    */
            /* And it will cause wake up failure. So the     */
            /* GPIO[2] will be used to control the           */
            /* Ramses reset pin to avoid the wake up problem */
            spr_data = MAC_IO_READ(MAC_SPR_REG);
            gpio_data = MAC_IO_READ(MAC_GPIO_REG);
            gpio_data &= 0xfffffL;      /* Make sure GPC bit is not set */
            if (spr_data & SPR_FRCS)
                gpio_data &= ~GPIO2_HI;     /* Set GPIO[2] to low */
            else
                gpio_data |= GPIO2_HI;

            MAC_IO_WRITE(MAC_GPIO_REG, gpio_data);
        }
    }
    return;
}

void
RamseD0Set (PHY *phy)
{
    if (phy->hpna_port == PHY_PORT_HPNA_EXT && phy->gpio_spi_phy_csn == 0)
    {
        /* Detected Ramses */
        if (phy->hw_options & PHY_HWOPTS_GP0_SPI_DOUT) {
            DWORD       gpio_data;
            /* Don't need GPIO[2] to control the Ramses reset pin */
            gpio_data = MAC_IO_READ(MAC_GPIO_REG);
            gpio_data &= 0xfffffL;      /* Make sure GPC bit is not set */
            gpio_data &= ~GPIO2_HI;     /* Set GPIO[2] to low */
            MAC_IO_WRITE(MAC_GPIO_REG, gpio_data);
        }
    }
    return;
}
