/*************************************************************************************************
 *                         PVCS Version Control Information                                      *
 *                                                                                               *
 * $Header:   P:\d878\HomeLAN\Archives\NdisDriver\Spi.c_v   1.6   08 Sep 1999 13:22:44   wangmx  $                            
 *                                                                                               *
 ************************************************************************************************/

/*
 *  Module Name: spi.c
 *  Author: Joe Bonaker
 *
 *  Copyright  1998-1999 Conexant Systems, Inc.
 *  All rights reserved.
 *
 *  This software may be used and distributed according to the terms
 *  of the GNU Public License, incorporated herein
 */

/*
 *  Module Description: HomePNA Serial Peripheral Interface (SPI)
 *  primitives supported by underlying "mii.c" MII and adapter/OS
 *  services.
 *
 * $Name: v2.20 $
 * $Log:   P:\d878\HomeLAN\Archives\NdisDriver\Spi.c_v  $
 * 
 *    Rev 1.6   08 Sep 1999 13:22:44   wangmx
 * JC:
 * Fix problem which prevent HCF from reading
 * EEPROM when HLAN is loaded.
 * 
 *    Rev 1.5   07 Jun 1999 10:55:10   wangmx
 *  
 * 
 *    Rev 1.4   May 12 1999 12:06:50   wangmx
 *  
 * Revision 1.3  1999/02/16 03:23:48  bonakeaj
 * - Disabled remote PHY cmds, DOS SpiWrite bug, & misc. cleanup
 * Revision 1.2  1998/11/20 17:06:42  bonakeaj
 * - 11623 integrated HomeLAN PHY support
 * Revision 1.1  1998/10/02 19:24:38  bonakeaj
 * Initial revision
 **********************************************************************/

#include "spi.h"

/* SPI instruction opcodes */
#define SPI_CMD_ENABLE_WRITE    0x06
#define SPI_CMD_DISABLE_WRITE   0x04
#define SPI_CMD_READ            0x03
#define SPI_CMD_WRITE           0x02

/* MII OS-mapping functions */
#ifndef OS_NANOSECOND_WAIT
    #if (DRV_OS == NDIS)
        #define OS_NANOSECOND_WAIT(n)
    #else
        #define OS_NANOSECOND_WAIT(n)   \
        MICROSECOND_WAIT((DWORD) ((n) + 999) / 1000)
    #endif
#endif
#define NANOSECOND_WAIT         OS_NANOSECOND_WAIT

#define IO_READ(offset)         MiiIoRead(phy, (offset))
#define IO_WRITE(offset, val)   MiiIoWrite(phy, (offset), (val))

/* LANfinity MAC GPIO registers */
#define MAC_R15_GPIO_REG        0x78            /* MAC >= RSS 11617 */
#define MAC_R12_GPIO_REG        0x60            /* MAC <= RSS 11611 */
#define MAC_R15_GP0             0x00010000L

#define GPIO_SET(_bit)          GpioOutputWrite(phy, (_bit), TRUE)
#define GPIO_CLR(_bit)          GpioOutputWrite(phy, (_bit), FALSE)
/* Modified SPI_CSN_ASSERT()/SPI_CSN_DEASSERT() define to support 623/Ramses board */
#define SPI_CSN_ASSERT()                                \
    if (phy->hpna_port == PHY_PORT_HPNA_EXT) {          \
        if (phy->gpio_spi_phy_csn) {                    \
            GPIO_CLR(phy->gpio_spi_phy_csn);            \
        } else {                                        \
            IO_WRITE(regAddr, (regValue &= ~SPR_EECS)); \
        }                                               \
    } else {                                            \
        IO_WRITE(regAddr, (regValue |= SPR_HLCS));      \
    }
#define SPI_CSN_DEASSERT()                              \
    if (phy->hpna_port == PHY_PORT_HPNA_EXT) {          \
        if (phy->gpio_spi_phy_csn) {                    \
            GPIO_SET(phy->gpio_spi_phy_csn);            \
        } else {                                        \
            IO_WRITE(regAddr, (regValue |= SPR_EECS));  \
        }                                               \
    } else {                                            \
        IO_WRITE(regAddr, (regValue &= ~SPR_HLCS));     \
    }

/* "spi.c" Module-specific functions */
static void SpiWriteCmd(PHY     *phy, DWORD cmdValue, int numBits);
static void GpioOutputWrite(PHY *phy, DWORD gpioBitMask, BOOLEAN setBit);

/* 11623/11617 Register 9 (offset 0x48) LAN_SPR assignments */
#define MAC_SPR_REG 0x48        /* same as MAC_MDIO_REG */
#define SPR_SWEN    0x02000000L /* 11617 Smart flash Write ENable */
#define SPR_MDIO    0x000F0000L /* 11617/11623 SMI (MDC/MDIO) bits */
#define SPR_RR      0x00004000L /* 11617 Read SPR_ER/FR device */
#define SPR_WR      0x00002000L /* 11617 Write SPR_ER/FR device */ 
#define SPR_FR      0x00001000L /* 11617 parallel Flash ROM select */
#define SPR_ER      0x00000800L /* 11617 serial EEPROM select */
#define SPR_HLDO    SPR_ER      /* 11623 HLAN Data Out (read) */
#define SPR_HLDI    0x00000400L /* 11623 HLAN Data In (write) */
#define SPR_HLCLK   0x00000200L /* 11623 HLAN Clock */
#define SPR_HLCS    0x00000100L /* 11623 HLAN Chip Select */
#define SPR_FRCS    0x00000010L /* 11623 serial Flash ROM Chip Select */
#define SPR_EEDO    0x00000008L /* serial EEPROM Data Out (read) */
#define SPR_EEDI    0x00000004L /* serial EEPROM Data In  (write) */
#define SPR_EECLK   0x00000002L /* serial EEPROM Clock */
#define SPR_EECS    0x00000001L /* serial EEPROM Chip Select */

/***********************************************************************
 * SpiInit() - Establish SPI GPIO register and bit assignment
 **********************************************************************/
void
SpiInit(PHY      *phy,
        PHY_PORT  spi_port,
        DWORD     hwreset_options,
        DWORD     gpio_chipselect)
{
    phy->hpna_port = spi_port;
    if (spi_port == PHY_PORT_HPNA_INT) {
        return;         /* 11623 internal HomePNA device */
    }

    /* Use 11617/external HLAN SPI (serial EEPROM) signal bits */
    /* Issue optional SPI PHY hardware reset */
    if (hwreset_options & PHY_HWOPTS_HWRESET_MASK) { 
        phy->gpio_spi_phy_reset = hwreset_options;
        PHY_HWRESET(phy, hwreset_options);
    }

    /* Establish the SPI chip select (CSN) line */
    phy->gpio_spi_phy_csn = gpio_chipselect;

    /* PhyInit() will enable write access to the SPI device */
    // SpiWriteEnable(phy);
}

/***********************************************************************
 * SpiWriteEnable() - Issues a "SET WE" command to SPI slave device
 **********************************************************************/
void
SpiWriteEnable(PHY *phy)
{
    SpiWriteCmd(phy, SPI_CMD_ENABLE_WRITE, 8 /* bits */);
}

/***********************************************************************
 * SpiWriteDisable() - Issues a "CLEAR WE" command to SPI slave device
 **********************************************************************/
void
SpiWriteDisable(PHY *phy)
{
    SpiWriteCmd(phy, SPI_CMD_DISABLE_WRITE, 8 /* bits */);
}

/***********************************************************************
 * SpiWriteByte() - Write an 8-bit data BYTE to an SPI slave device
 **********************************************************************/
void
SpiWriteByte(PHY *phy, BYTE byteAddr, BYTE byteValue)
{
    DWORD       writeCmd;

    writeCmd = ((DWORD) SPI_CMD_WRITE << 16) |
               ((DWORD) byteAddr      <<  8) |
               byteValue;
    SpiWriteCmd(phy, writeCmd, 24 /* bits */);
}

/***********************************************************************
 * SpiWriteWord() - Write a 16-bit data WORD to an SPI slave device
 **********************************************************************/
void
SpiWriteWord(PHY *phy, BYTE byteAddr, WORD wordValue)
{
    SpiWriteByte(phy, byteAddr,              (BYTE) wordValue);
    SpiWriteByte(phy, (BYTE) (byteAddr + 1), (BYTE) (wordValue >> 8));
}

/***********************************************************************
 * SpiWriteDword() - Write a 32-bit data DWORD to an SPI slave device
 **********************************************************************/
void
SpiWriteDword(PHY *phy, BYTE byteAddr, DWORD dwordValue)
{
    SpiWriteWord(phy, byteAddr,              (WORD) dwordValue);
    SpiWriteWord(phy, (BYTE) (byteAddr + 2), (WORD) (dwordValue >> 16));
}

/***********************************************************************
 * SpiReadByte() - Read an 8-bit data BYTE from an SPI slave device
 **********************************************************************/
BYTE
SpiReadByte(PHY *phy, BYTE byteAddr)
{
    BYTE        dataValue;
    WORD        regAddr = MAC_SPR_REG;
    DWORD       regValue;
    WORD        readCmd;
    DWORD       sdout;          /* SPI data out => host input */
    DWORD       sdin;           /* SPI data in  => host output */ 
    DWORD       sclk;           /* SPI clock    => host output */
    DWORD       n;

    if (phy->hpna_port == PHY_PORT_NONE) {
        return (0x00);
    }

    if (phy->hpna_port == PHY_PORT_HPNA_INT) {
        /* Use 11623 internal HLAN SPI signal bits */
        sdout  = SPR_HLDO;
        sdin   = SPR_HLDI;
        sclk   = SPR_HLCLK;
    } else {
        /* Use 11617/external HLAN SPI (serial EEPROM) signal bits */
        sdout  = SPR_EEDO;
        sdin   = SPR_EEDI;
        sclk   = SPR_EECLK;
    }

    /* Preserve ROM_CS# (Ramses RESET#) and SMI (MDC/MDIO) bits */
    regValue = IO_READ(regAddr) & (SPR_FRCS | SPR_MDIO);

    /* Configure CSR9, LAN Serial Port Register (SPR) */
    regValue |= (SPR_ER | SPR_WR);      /* 11617 compatibility */
    IO_WRITE(regAddr, regValue);
    NANOSECOND_WAIT(50);                /* TSETcsh */

    /* Assert chip select */
    SPI_CSN_ASSERT();
    NANOSECOND_WAIT(50);                /* TSETcsl */

    /* Shift out (to SI) the 8-bit SPI opcode and 8-bit address ... */
    readCmd = (SPI_CMD_READ << 8) | byteAddr;
    for (n = 0x8000; n != 0; n >>= 1) {
        regValue &= ~sdin;
        if (readCmd & n) {
            regValue |= sdin;    /* '1' bit output via SI */
        }
        IO_WRITE(regAddr, regValue);
        NANOSECOND_WAIT(400);           /* TSETsiv, TWIDTHclkl */

        IO_WRITE(regAddr, regValue | sclk);
        NANOSECOND_WAIT(400);           /* TWIDTHclkh */
    }

    /* Drop <sclk> and switch SPR mode from write to read */
    regValue &= ~(SPR_WR | sdin);
    regValue |= SPR_RR;
    IO_WRITE(regAddr, regValue);
    NANOSECOND_WAIT(50);                /* TDLYsov */
    
    /* ... and shift in (from SO) the 8-bit SPI data byte */
    for (n = 0, dataValue = 0; n < 8; n++) {
        dataValue <<= 1;
        if (phy->hw_options & PHY_HWOPTS_GP0_SPI_DOUT) {
            if (IO_READ(MAC_R15_GPIO_REG) & MAC_R15_GP0) {
                dataValue++;
            }
        } else if (IO_READ(regAddr) & sdout) {
            dataValue++;
        }
        IO_WRITE(regAddr, regValue | sclk);
        NANOSECOND_WAIT(400);           /* TWIDTHclkh */

        IO_WRITE(regAddr, regValue);
        NANOSECOND_WAIT(400);           /* TDLYsov, TWIDTHclkl */
    }

    /* Remove chip select */
    SPI_CSN_DEASSERT();
    NANOSECOND_WAIT(50);                /* TSETcsl */

    /* Assert chip select so that eeprom read/write can be function properly */
    SPI_CSN_ASSERT();

    return (dataValue);
}

/***********************************************************************
 * SpiReadWord() - Read a 16-bit data WORD from an SPI slave device
 **********************************************************************/
WORD
SpiReadWord(PHY *phy, BYTE byteAddr)
{
    WORD        wordValue;

    wordValue =    (WORD) SpiReadByte(phy, byteAddr);
    wordValue += (((WORD) SpiReadByte(phy,
                                      (BYTE) (byteAddr + 1))) << 8);

    return (wordValue);
}

/***********************************************************************
 * SpiReadDword() - Read a 32-bit data DWORD from an SPI slave device
 **********************************************************************/
DWORD
SpiReadDword(PHY *phy, BYTE byteAddr)
{
    DWORD       dwordValue;

    dwordValue =    (DWORD) SpiReadWord(phy, byteAddr);
    dwordValue += (((DWORD) SpiReadWord(phy,
                                        (BYTE) (byteAddr + 2))) << 16);

    return (dwordValue);
}

/***********************************************************************
 * GpioOutputWrite() - Sets or clears one or more GPIO output signals
 **********************************************************************/
static void
GpioOutputWrite(PHY *phy, DWORD gpioBitMask, BOOLEAN setBit)
{
    WORD        gpioRegAddr = MAC_R15_GPIO_REG;
    DWORD       gpioValue;

    gpioBitMask &= LAN_GPIO_MASK;
    if (phy->r12_gpio) {
        /* some pre-production devices use R12 for GPIO */
        gpioRegAddr = MAC_R12_GPIO_REG;
        gpioBitMask >>= 16; /* R15 bits [19:16] => R12 bits [3:0] */
    }

    gpioValue = IO_READ(gpioRegAddr);
    if (setBit) {
        gpioValue |= gpioBitMask;
    } else {
        gpioValue &= ~gpioBitMask;
    }
    IO_WRITE(gpioRegAddr, gpioValue);
}

/***********************************************************************
 * SpiWriteCmd() - Issues WRITE/WEN/WDS instruction to SPI device
 **********************************************************************/
static void
SpiWriteCmd(PHY    *phy, DWORD cmdValue, int numBits)
{
    WORD        regAddr = MAC_SPR_REG;
    DWORD       regValue;
    DWORD       sdin;           /* SPI data in  => host output */ 
    DWORD       sclk;           /* SPI clock    => host output */
    DWORD       n;

    if (phy->hpna_port == PHY_PORT_HPNA_INT) {
        /* Use 11623 internal HLAN SPI signal bits */
        sdin   = SPR_HLDI;
        sclk   = SPR_HLCLK;
    } else {
        /* Use 11617/external HLAN SPI (serial EEPROM) signal bits */
        sdin   = SPR_EEDI;
        sclk   = SPR_EECLK;
    }

    /* Suppress in read-only (debug) mode or when device not present */
    if (phy->read_only || !(phy->devices & phy->hpna_port)) {
        return;
    }

    /* Preserve ROM_CS# (Ramses RESET#) and SMI (MDC/MDIO) bits */
    regValue = IO_READ(regAddr) & (SPR_FRCS | SPR_MDIO);

    /* Configure CSR9, LAN Serial Port Register (SPR) */
    regValue |= (SPR_ER | SPR_WR);      /* 11617 compatibility */
    IO_WRITE(regAddr, regValue);
    NANOSECOND_WAIT(50);                /* TSETcsh */

    /* Issue chip select */
    SPI_CSN_ASSERT();
    NANOSECOND_WAIT(50);                /* TSETcsl */

    /* Shift out (to SI) the 8-bit SPI opcode and optional addr/data */
    for (n = (DWORD) 0x01 << (numBits - 1); n != 0; n >>= 1) {
        regValue &= ~sdin;
        if (cmdValue & n) {
            regValue |= sdin;    /* '1' bit output via SI */
        }
        IO_WRITE(regAddr, regValue);
        NANOSECOND_WAIT(400);           /* TSETsiv, TWIDTHclkl */

        IO_WRITE(regAddr, regValue | sclk);
        NANOSECOND_WAIT(400);           /* TWIDTHclkh */
    }

    /* Drop <sclk> and switch SPR mode from write to read */
    regValue &= ~(SPR_WR | sdin);
    regValue |= SPR_RR;
    IO_WRITE(regAddr, regValue);
    NANOSECOND_WAIT(100);               /* TDLYsoz */
    
    /* Remove chip select */
    SPI_CSN_DEASSERT();
    NANOSECOND_WAIT(50);                /* TSETcsl */

    /* Assert chip select so that eeprom read/write can be function properly */
    SPI_CSN_ASSERT();
}
